from assemblyline import odm

MSG_TYPES = {"AlerterHeartbeat"}
LOADER_CLASS = "assemblyline.odm.messages.alerter_heartbeat.AlerterMessage"


@odm.model()
class Queues(odm.Model):
    alert = odm.Integer()


@odm.model()
class Metrics(odm.Model):
    created = odm.Integer()
    error = odm.Integer()
    received = odm.Integer()
    updated = odm.Integer()


@odm.model()
class Heartbeat(odm.Model):
    instances = odm.Integer()
    metrics = odm.Compound(Metrics)
    queues = odm.Compound(Queues)


@odm.model()
class AlerterMessage(odm.Model):
    msg = odm.Compound(Heartbeat)
    msg_loader = odm.Enum(values={LOADER_CLASS}, default=LOADER_CLASS)
    msg_type = odm.Enum(values=MSG_TYPES, default="AlerterHeartbeat")
    sender = odm.Keyword()



import hashlib
import uuid

import baseconv

TINY = 8
SHORT = 16
MEDIUM = NORMAL = 32
LONG = 64


def get_random_id():
    return baseconv.base62.encode(uuid.uuid4().int)


def get_id_from_data(data, prefix=None, length=MEDIUM):
    possible_len = [TINY, SHORT, MEDIUM, LONG]
    if length not in possible_len:
        raise ValueError(f"Invalid hash length of {length}. Possible values are: {str(possible_len)}.")
    sha256_hash = hashlib.sha256(str(data).encode()).hexdigest()[:length]
    _hash = baseconv.base62.encode(int(sha256_hash, 16))

    if isinstance(prefix, str):
        _hash = f"{prefix}_{_hash}"

    return _hash



import collections
import logging
import pprint
import threading

# use pytz (as demanded by apschedular) to ensure the timezone is set,
# boxes with weird/new/different default timezone values can break things
# if a value APS likes isn't set explicitly
import pytz

from assemblyline.common import forge
from assemblyline.common.uid import get_random_id
from assemblyline.odm.messages import PerformanceTimer
from assemblyline.remote.datatypes import get_client

log = logging.getLogger('assemblyline.counters')


# noinspection PyAbstractClass
class Counters(collections.Counter):
    pass


# noinspection PyBroadException
class AutoExportingCounters(object):
    """
    A wrapper around collections.Counter that adds periodic backup.

    NOTE: Not to be confused with remote_datatypes. Counter which wraps a live
          redis counter. This counter is save only, and only offers weak durability.
          This is appropriate for monitoring and performance measurements, not
          for operational counters that require strict semantics.

    At the specified interval and program exit, the value in the counters will be
    sent to the provided channel.
    """

    def __init__(self,
                 name,
                 host=None,
                 export_interval_secs=None,
                 counter_type=None,
                 config=None,
                 redis=None,
                 counter_names=None,
                 timer_names=None):
        config = config or forge.get_config()
        self.channel = forge.get_metrics_sink(redis)
        self.export_interval = export_interval_secs or config.core.metrics.export_interval
        self.name = name
        self.host = host or get_random_id()
        self.type = counter_type or name

        self.counter_schema = set(counter_names)
        self.timer_schema = set(timer_names)

        self.counts = None
        self.lock = threading.Lock()
        self.scheduler = None
        self.reset()

        assert self.channel
        assert(self.export_interval > 0)

    # noinspection PyUnresolvedReferences
    def start(self):
        from apscheduler.schedulers.background import BackgroundScheduler
        import atexit

        self.scheduler = BackgroundScheduler(daemon=True, timezone=pytz.utc)
        self.scheduler.add_job(self.export, 'interval', seconds=self.export_interval)
        self.scheduler.start()

        atexit.register(lambda: self.stop())

    def reset(self):
        with self.lock:
            old, self.counts = self.counts, Counters({key: 0 for key in self.counter_schema})
            self.counts.update({key + '.t': 0 for key in self.timer_schema})
            self.counts.update({key + '.c': 0 for key in self.timer_schema})
            self.counts['type'] = self.type
            self.counts['name'] = self.name
            self.counts['host'] = self.host

        return old

    def stop(self):
        if self.scheduler:
            self.scheduler.shutdown(wait=False)
            self.scheduler = None
        self.export()

    def export(self):
        try:
            # To avoid blocking increments on the redis operation
            # we only hold the long to do a copy.
            thread_copy = dict(self.reset().items())
            self.channel.publish(thread_copy)
            log.debug(f"{pprint.pformat(thread_copy)}")

            return thread_copy
        except Exception:
            log.exception("Exporting counters")

    def increment(self, name, increment_by=1):
        try:
            if name not in self.counter_schema:
                raise ValueError(f"{name} is not an accepted counter for this module: f{self.counter_schema}")
            with self.lock:
                self.counts[name] += increment_by
                return increment_by
        except Exception:  # Don't let increment fail anything.
            log.exception("Incrementing counter")
            return 0

    def increment_execution_time(self, name, execution_time):
        try:
            if name not in self.timer_schema:
                raise ValueError(f"{name} is not an accepted counter for this module: f{self.timer_schema}")
            with self.lock:
                self.counts[name + ".c"] += 1
                self.counts[name + ".t"] += execution_time
                return execution_time
        except Exception:  # Don't let increment fail anything.
            log.exception("Incrementing counter")
            return 0


def export_metrics_once(name, schema, metrics, host=None, counter_type=None, config=None, redis=None):
    """Manually publish metric counts to the metrics system.

    This was built for when the service server is reporting metrics for execution and caching
    on behalf of many services. At the moment the metrics system uses the hosts to count the number
    of instances of each service. This could be done with a single auto exporting counter for
    the service server, but that may require significant downstream changes in the metrics system.
    """
    config = config or forge.get_config()
    redis = redis or get_client(
        config.core.metrics.redis.host,
        config.core.metrics.redis.port,
        False
    )

    # Separate out the timers and normal counters
    timer_schema = set()
    counter_schema = set()

    for _k, field_type in schema.fields().items():
        if isinstance(field_type, PerformanceTimer):
            timer_schema.add(_k)
        else:
            counter_schema.add(_k)

    for _k in timer_schema:
        counter_schema.discard(_k + '_count')

    channel = forge.get_metrics_sink(redis)

    counts = Counters({key: 0 for key in counter_schema})
    counts.update({key + '.t': 0 for key in timer_schema})
    counts.update({key + '.c': 0 for key in timer_schema})

    for metric, value in metrics.items():
        if metric in counter_schema:
            counts[metric] += value
        elif metric in timer_schema:
            counts[name + ".c"] += 1
            counts[name + ".t"] += value
        else:
            raise ValueError(f"{metric} is not an accepted counter")

    counts['type'] = counter_type
    counts['name'] = name
    counts['host'] = host

    channel.publish(dict(counts.items()))



import re

from typing import List, Dict

from assemblyline.common.dict_utils import flatten
from assemblyline.odm.models.tagging import Tagging


def tag_list_to_dict(tag_list: List[Dict]) -> Dict:
    tag_dict = {}
    for t in tag_list:
        if t['type'] not in tag_dict:
            tag_dict[t['type']] = []
        tag_dict[t['type']].append(t['value'])

    return tag_dict


def tag_dict_to_list(tag_dict: Dict) -> List[Dict]:
    return [
        {'type': k, 'value': t, 'short_type': k.rsplit(".", 1)[-1]}
        for k, v in flatten(tag_dict).items()
        if v is not None
        for t in v
    ]


class InvalidWhitelist(Exception):
    pass


class TagWhitelister(object):
    def __init__(self, data, log=None):
        valid_tags = set(Tagging.flat_fields().keys())

        self.match = data.get('match', {})
        self.regex = data.get('regex', {})
        self.log = log

        # Validate matches and regex
        for section, item in {'match': self.match, 'regex': self.regex}.items():
            if not isinstance(item, dict):
                raise InvalidWhitelist(f"Section {section} should be of type: DICT")

            for k, v in item.items():
                if not isinstance(v, list):
                    raise InvalidWhitelist(f"Values in the {section} section should all be of type: LIST")

                if k not in valid_tags:
                    raise InvalidWhitelist(f"Key ({k}) in the {section} section is not a valid tag.")

                if section == 'regex':
                    self.regex[k] = [re.compile(x) for x in v]

    def is_whitelisted(self, t_type, t_value):
        for match in self.match.get(t_type, []):
            if t_value == match:
                if self.log:
                    self.log.info(f"Tag '{t_type}' with value '{t_value}' was whitelisted by match rule.")
                return True

        for regex in self.regex.get(t_type, []):
            if regex.match(t_value):
                if self.log:
                    self.log.info(f"Tag '{t_type}' with value '{t_value}' "
                                  f"was whitelisted by regex '{regex.pattern}'.")
                return True

        return False

    def whitelist_many(self, t_type, t_values):
        if not isinstance(t_values, list):
            t_values = [t_values]
        return [x for x in t_values if not self.is_whitelisted(t_type, x)]

    def get_validated_tag_map(self, tag_map):
        return {k: self.whitelist_many(k, v) for k, v in tag_map.items() if v is not None}



import logging
import os
import shutil

from assemblyline.common.exceptions import ChainAll
from assemblyline.common.uid import get_random_id
from assemblyline.filestore.transport.base import Transport, TransportException, normalize_srl_path


@ChainAll(TransportException)
class TransportLocal(Transport):
    """
    Local file system Transport class.
    """

    def __init__(self, base=None, normalize=None):
        self.log = logging.getLogger('assemblyline.transport.local')
        self.base = base
        self.host = "localhost"

        def local_normalize(path):
            # If they've provided an absolute path. Leave it a is.
            if path.startswith('/'):
                s = path
            # Relative paths
            elif '/' in path or len(path) != 64:
                s = _join(self.base, path)
            else:
                s = _join(self.base, normalize_srl_path(path))
            self.log.debug('local normalized: %s -> %s', path, s)
            return s

        if not normalize:
            normalize = local_normalize

        super(TransportLocal, self).__init__(normalize=normalize)

    def delete(self, path):
        path = self.normalize(path)
        os.unlink(path)

    def exists(self, path):
        path = self.normalize(path)
        return os.path.exists(path)

    def getmtime(self, path):
        path = self.normalize(path)

        try:
            return os.path.getmtime(path)
        except OSError:
            return 0

    def makedirs(self, path):
        path = self.normalize(path)
        try:
            os.makedirs(path)
        except OSError as e:
            if e.errno == 17:
                pass
            else:
                raise e

    # File based functions
    def download(self, src_path, dst_path):
        if src_path == dst_path:
            return

        src_path = self.normalize(src_path)
        dir_path = os.path.dirname(dst_path)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        shutil.copy(src_path, dst_path)

    def upload(self, src_path, dst_path):
        dst_path = self.normalize(dst_path)
        if src_path == dst_path:
            return

        dirname = os.path.dirname(dst_path)
        filename = os.path.basename(dst_path)
        tempname = get_random_id()
        temppath = _join(dirname, tempname)
        finalpath = _join(dirname, filename)
        assert (finalpath == dst_path)
        self.makedirs(dirname)
        shutil.copy(src_path, temppath)
        shutil.move(temppath, finalpath)
        assert (self.exists(dst_path))

    # Buffer based functions
    def get(self, path):
        path = self.normalize(path)
        fh = None
        try:
            fh = open(path, "rb")
            return fh.read()
        finally:
            if fh:
                fh.close()

    def put(self, path, content):
        path = self.normalize(path)

        dirname = os.path.dirname(path)
        filename = os.path.basename(path)

        tempname = get_random_id()
        temppath = _join(dirname, tempname)

        finalpath = _join(dirname, filename)
        assert(finalpath == path)

        self.makedirs(dirname)
        fh = None
        try:
            fh = open(temppath, "wb")
            return fh.write(content)
        finally:
            if fh:
                fh.close()

            try:
                shutil.move(temppath, finalpath)
            except shutil.Error:
                pass
            assert(self.exists(path))

    def __str__(self):
        return 'file://{}'.format(self.base)

###############################
# Helper functions.
###############################


def _join(base, path):
    path = path.replace("\\", "/").replace("//", "/")
    if base is None:
        return path
    return os.path.join(base, path.lstrip("/")).replace("\\", "/")



from assemblyline.common.exceptions import ChainException


def normalize_srl_path(srl):
    if '/' in srl:
        return srl

    return '{0}/{1}/{2}/{3}/{4}'.format(srl[0], srl[1], srl[2], srl[3], srl)


class TransportException(ChainException):
    """
    FileTransport exception base class.
    
    TransportException is a subclass of ChainException so that it can be
    used with the Chain and ChainAll decorators.
    """
    pass


class Transport(object):
    """
    FileTransport base class.
    
    - Subclasses should override all methods.
    - Except as noted, FileTransport methods do not return value and raise
    - TransportException on failure.
    - Methods should only raise TransportExceptions. (The decorators
      Chain and ChainAll can be applied to a function/method and class,
      respectively, to ensure that any exceptions raised are converted to
      TransportExceptions.
    """
    
    def __init__(self, normalize=normalize_srl_path):
        self.normalize = normalize

    def close(self):
        pass
    
    def delete(self, path):
        """
        Deletes the file.
        """
        raise TransportException("Not Implemented")

    def exists(self, path):
        """
        Returns True if the path exists, False otherwise.
        Should work with both files and directories.
        """
        raise TransportException("Not Implemented")

    def makedirs(self, path):
        """
        Like os.makedirs the super-mkdir, create the leaf directory path and
        any intermediate path segments.
        """
        raise TransportException("Not Implemented")

    # File based functions
    def download(self, src_path, dst_path):
        """
        Copies the content of the filestore src_path to the local dst_path.
        """
        raise TransportException("Not Implemented")

    def upload(self, src_path, dst_path):
        """
        Save upload source file src_path to to the filesotre dst_path, overwriting dst_path if it already exists.
        """
        raise TransportException("Not Implemented")

    def upload_batch(self, local_remote_tuples):
        """
        Upload multiple files specified by list of (local, remote) tuples.
        Transports that can optimize batch file transfers should write a custom upload_batch.
        """
        failed_tuples = []
        for (src_path, dst_path) in local_remote_tuples:
            try:
                self.upload(src_path, dst_path)
            except Exception as e:
                failed_tuples.append((src_path, dst_path, str(e)))
        return failed_tuples

    # Buffer based functions
    def get(self, path):
        """
        Returns the content of the file.
        """
        raise TransportException("Not Implemented")
        
    def put(self, dst_path, content):
        """
        Put the content of the file in memory directly to the filestore dst_path
        """
        raise TransportException("Not Implemented")



from assemblyline import odm


@odm.model(index=True, store=True)
class Seen(odm.Model):
    count = odm.Integer(default=1)      # Number of time we've seen this file
    first = odm.Date(default="NOW")     # Date of the first time we've seen the file
    last = odm.Date(default="NOW")      # Date of the last time we've seen the file


@odm.model(index=True, store=True)
class File(odm.Model):
    archive_ts = odm.Date(store=False)                # Archiving timestamp
    ascii = odm.Keyword(index=False, store=False)     # Dotted ascii representation of the first 64 bytes of the file
    classification = odm.Classification()             # Classification of the file
    entropy = odm.Float()                             # Entropy of the file
    expiry_ts = odm.Optional(odm.Date(store=False))   # Expiry timestamp
    hex = odm.Keyword(index=False, store=False)       # Hex dump of the first 64 bytes of the file
    md5 = odm.MD5(copyto="__text__")                  # MD5 of the top level file
    magic = odm.Keyword(store=False)                  # Output from libmagic related to that file
    mime = odm.Optional(odm.Keyword(store=False))     # Mime type of the file as identified by libmagic
    seen = odm.Compound(Seen, default={})             # Attributes about when the file was seen
    sha1 = odm.SHA1(copyto="__text__")                # SHA1 hash of the file
    sha256 = odm.SHA256(copyto="__text__")            # SHA256 hash of the file
    size = odm.Integer()                              # Size of the file
    ssdeep = odm.SSDeepHash(store=False)              # SSDEEP hash of the file
    type = odm.Keyword(copyto="__text__")             # Type of file as identified by Assemblyline



# TODO: Are we still using this?


class Context(object):

    AFFECTS = 'Affects'
    COMPROMISED = 'Compromised'
    DERIVED_FROM = 'Derived From'
    DYNAMIC = 'Dynamic'
    IDENTIFIES = 'Identifies'
    STATIC = 'Static'
    PART_OF = 'Part Of'
    RESPONSIBLE_FOR = 'Responsible For'
    VARIANT_OF = 'Variant Of'
    COMMON_CONTEXT = [AFFECTS, COMPROMISED, DERIVED_FROM, DYNAMIC, IDENTIFIES, PART_OF, RESPONSIBLE_FOR, STATIC,
                      VARIANT_OF]

    BEACONS = 'Beacons'
    CONNECTS_TO = 'Connects To'
    DOWNLOADS = 'Downloads'
    LINKED_TO = 'Linked To'
    RESOLVES_TO = 'Resolves To'
    SENDS_TO = 'Sends To'
    SENT_AS_BCC_TO = 'Sent As BCC To'
    SENT_AS_CC_TO = 'Sent As CC To'
    SPOOFER_OF = 'Spoofer Of'
    NETWORKING_CONTEXT = [BEACONS, CONNECTS_TO, DOWNLOADS, LINKED_TO, RESOLVES_TO, SENDS_TO, SENT_AS_BCC_TO,
                          SENT_AS_CC_TO, SPOOFER_OF]

    CHILD_OF = 'Child Of'
    CONTAINS = 'Contains'
    OWNER_OF = 'Owner Of'
    USER_OF = 'User Of'
    FILE_CONTEXT = [CHILD_OF, CONTAINS, OWNER_OF, USER_OF]

    # List of allowed Tag to Context mappings
    # If not explicitly specified, COMMON_CONTEXT is defaulted for the tag
    RECOGNIZED_CONTEXT = {
        'AUTORUN_': COMMON_CONTEXT + FILE_CONTEXT,
        'DYNAMIC_': COMMON_CONTEXT + FILE_CONTEXT,
        'FILE_': COMMON_CONTEXT + FILE_CONTEXT,
        'FILENAME_': COMMON_CONTEXT + FILE_CONTEXT,
        'NET_': COMMON_CONTEXT + NETWORKING_CONTEXT,
        'PE_': COMMON_CONTEXT + FILE_CONTEXT,
        'REGISTRY_': COMMON_CONTEXT + FILE_CONTEXT,
        'SERVICE_': None,
        'TECHNIQUE_': COMMON_CONTEXT + FILE_CONTEXT + NETWORKING_CONTEXT,
        'HEURISTIC': None,
        'REQUEST_USERNAME': None,
        'REQUEST_SCORE': None,
        'DISPLAY_STRING_SEARCH': None,
    }

    STARTWITH_CONTEXT = {
        'AV_': ["scanner:"]
    }

    @staticmethod
    def verify_context(tag_type, test_context):
        # Check Recognized Context pairs first
        for key in Context.RECOGNIZED_CONTEXT.keys():
            if tag_type.startswith(key) and test_context in Context.RECOGNIZED_CONTEXT[key]:
                return True

        # Default to Common Context
        if test_context in Context.COMMON_CONTEXT:
            return True

        for key in Context.STARTWITH_CONTEXT.keys():
            if tag_type.startswith(key):
                for item in Context.STARTWITH_CONTEXT[key]:
                    if test_context.startswith(item):
                        return True

        return False



from assemblyline import odm
from assemblyline.common import forge
Classification = forge.get_classification()


@odm.model(index=False, store=False)
class Favorite(odm.Model):
    created_by = odm.Keyword()                           # Who created the favorite
    classification = odm.Classification(
        is_user_classification=True, copyto="__text__",
        default=Classification.UNRESTRICTED)             # Classification of the favorite
    name = odm.Keyword()                                 # Name of the favorite
    query = odm.Keyword()                                # Query for the favorite


@odm.model(index=False, store=False)
class UserFavorites(odm.Model):
    alert = odm.List(odm.Compound(Favorite), default=[])       # Alert page favorites
    error = odm.List(odm.Compound(Favorite), default=[])       # Error page favorites
    search = odm.List(odm.Compound(Favorite), default=[])      # Search page favorites
    signature = odm.List(odm.Compound(Favorite), default=[])   # Signature page favorites
    submission = odm.List(odm.Compound(Favorite), default=[])  # Submission page favorites



import json

from assemblyline.remote.datatypes import get_client, retry_call, decode

# Work around for inconsistency between ZRANGEBYSCORE and ZREMRANGEBYSCORE
#   (No limit option available or we would just be using that directly)
#
# args:
#   minimum score to pop
#   maximum score to pop
#   number of elements to skip before popping any
#   max element count to pop
pq_dequeue_range_script = """
local unpack = table.unpack or unpack
local min_score = tonumber(ARGV[1]); 
if min_score == nil then min_score = -math.huge end
local max_score = tonumber(ARGV[2]);
if max_score == nil then max_score = math.huge end 
local rem_offset = tonumber(ARGV[3]); 
local rem_limit = tonumber(ARGV[4]); 

local entries = redis.call("zrangebyscore", KEYS[1], -max_score, -min_score, "limit", rem_offset, rem_limit);
if #entries > 0 then redis.call("zrem", KEYS[1], unpack(entries)) end
return entries 
"""


# ARGV[1]: <queue name>
# ARGV[2]: <max items to pop minus one>
pq_pop_script = """
local result = redis.call('zrange', ARGV[1], 0, ARGV[2])
if result then redis.call('zremrangebyrank', ARGV[1], 0, ARGV[2]) end
return result
"""

# ARGV[1]: <queue name>
# ARGV[2]: <priority>
# ARGV[3]: <vip>,
# ARGV[4]: <item (string) to push>
pq_push_script = """
local seq = string.format('%020d', redis.call('incr', 'global-sequence'))
local vip = string.format('%1d', ARGV[3])
local value = vip..seq..ARGV[4]
redis.call('zadd', ARGV[1], 0 - ARGV[2], value)
return value
"""

# ARGV[1]: <queue name>
# ARGV[2]: <max items to unpush>
pq_unpush_script = """
local result = redis.call('zrange', ARGV[1], 0 - ARGV[2], 0 - 1)
if result then redis.call('zremrangebyrank', ARGV[1], 0 - ARGV[2], 0 - 1) end
return result
"""


class PriorityQueue(object):
    def __init__(self, name, host=None, port=None, private=False):
        self.c = get_client(host, port, private)
        self.r = self.c.register_script(pq_pop_script)
        self.s = self.c.register_script(pq_push_script)
        self.t = self.c.register_script(pq_unpush_script)
        self._deque_range = self.c.register_script(pq_dequeue_range_script)
        self.name = name

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.delete()

    def count(self, lowest, highest):
        return retry_call(self.c.zcount, self.name, -highest, -lowest)

    def delete(self):
        retry_call(self.c.delete, self.name)

    def length(self):
        return retry_call(self.c.zcard, self.name)

    def pop(self, num=None):
        if num is not None and num <= 0:
            return []

        if num:
            return [decode(s[21:]) for s in retry_call(self.r, args=[self.name, num-1])]
        else:
            ret_val = retry_call(self.r, args=[self.name, 0])
            if ret_val:
                return decode(ret_val[0][21:])
            return None

    def blocking_pop(self, timeout=0):
        """When only one item is requested, blocking is is possible."""
        result = retry_call(self.c.bzpopmin, self.name, timeout)
        if result:
            return decode(result[1][21:])
        return None

    def dequeue_range(self, lower_limit='', upper_limit='', skip=0, num=1):
        """Dequeue a number of elements, within a specified range of scores.

        Limits given are inclusive, can be made exclusive, see redis docs on how to format limits for that.

        NOTE: lower/upper limit is negated+swapped in the lua script, no need to do it here

        :param lower_limit: The score of all dequeued elements must be higher or equal to this.
        :param upper_limit: The score of all dequeued elements must be lower or equal to this.
        :param skip: In the range of available items to dequeue skip over this many.
        :param num: Maximum number of elements to dequeue.
        :return: list
        """
        results = retry_call(self._deque_range, keys=[self.name], args=[lower_limit, upper_limit, skip, num])
        return [decode(res[21:]) for res in results]

    def push(self, priority: int, data, vip=None):
        vip = 0 if vip else 9
        return retry_call(self.s, args=[self.name, priority, vip, json.dumps(data)])

    def rank(self, raw_value):
        return retry_call(self.c.zrank, self.name, raw_value)

    def remove(self, raw_value):
        return retry_call(self.c.zrem, self.name, raw_value)

    def unpush(self, num=None):
        if num is not None and num <= 0:
            return []

        if num:
            return [decode(s[21:]) for s in retry_call(self.t, args=[self.name, num])]
        else:
            ret_val = retry_call(self.t, args=[self.name, 1])
            if ret_val:
                return decode(ret_val[0][21:])
            return None


class UniquePriorityQueue(PriorityQueue):
    """A priority queue where duplicate entries will be merged."""

    def __init__(self, name, host=None, port=None, private=False):
        super().__init__(name, host, port, private)
        del self.s

    def remove(self, data: str):
        """Remove a value from the priority  queue."""
        retry_call(self.c.zrem, self.name, json.dumps(data))

    def push(self, priority: int, data, vip=None) -> int:
        """Add or update elements in the priority queue.

        Existing elements will have their priority updated.

        Returns:
            Number of _NEW_ elements in the queue after the operation.
        """
        return retry_call(self.c.zadd, self.name, {json.dumps(data): -priority})

    def pop(self, num=None):
        if num is not None and num <= 0:
            return []

        if num:
            return [decode(s) for s in retry_call(self.r, args=[self.name, num-1])]
        else:
            ret_val = retry_call(self.r, args=[self.name, 0])
            if ret_val:
                return decode(ret_val[0])
            return None

    def unpush(self, num=None):
        if num is not None and num <= 0:
            return []

        if num:
            return [decode(s) for s in retry_call(self.t, args=[self.name, num])]
        else:
            ret_val = retry_call(self.t, args=[self.name, 1])
            if ret_val:
                return decode(ret_val[0])
            return None

    def dequeue_range(self, lower_limit='', upper_limit='', skip=0, num=1):
        """Dequeue a number of elements, within a specified range of scores.

        Limits given are inclusive, can be made exclusive, see redis docs on how to format limits for that.

        NOTE: lower/upper limit is negated+swapped in the lua script, no need to do it here

        :param lower_limit: The score of all dequeued elements must be higher or equal to this.
        :param upper_limit: The score of all dequeued elements must be lower or equal to this.
        :param skip: In the range of available items to dequeue skip over this many.
        :param num: Maximum number of elements to dequeue.
        :return: list
        """
        results = retry_call(self._deque_range, keys=[self.name], args=[lower_limit, upper_limit, skip, num])
        return [decode(res) for res in results]


def select(*queues, **kw):
    timeout = kw.get('timeout', 0)
    if len(queues) < 1:
        raise TypeError('At least one queue must be specified')
    if any([type(q) != PriorityQueue for q in queues]):
        raise TypeError('Only NamedQueues supported')

    c = queues[0].c
    response = retry_call(c.bzpopmin, [q.name for q in queues], timeout)

    if not response:
        return response

    return response[0].decode('utf-8'), json.loads(response[1][21:])



import json
import os
import random
import requests
import string
import time

from requests.exceptions import ConnectionError


class SetupException(Exception):
    pass


def set_password(user, password):
    try:
        print(f"\nSetting up {user} password...")
        req = requests.post(f"http://elastic:{ELASTIC_PASSWORD}@{ELASTIC_HOST}:9200/_security/user/{user}/_password",
                            headers={"Content-Type": "application/json"},
                            data=json.dumps({"password": password}))
        if not req.ok:
            raise SetupException(f"ERROR: Failed to set password for user: {user}")

        print("Password set")
    except SetupException as e:
        print(str(e))


def create_role(name, cluster_priv=None, indices=None):
    try:
        print(f"\nCreating role {name}...")

        if cluster_priv is None:
            cluster_priv = []

        if indices is None:
            indices = []

        role = {"cluster": cluster_priv,
                "indices": [
                    {"names": indices,
                     "privileges": ["all"],
                     "allow_restricted_indices": False}],
                "applications": [],
                "run_as": [],
                "metadata": {}}
        req = requests.post(f"http://elastic:{ELASTIC_PASSWORD}@{ELASTIC_HOST}:9200/_security/role/{name}",
                            headers={"Content-Type": "application/json"},
                            data=json.dumps(role))

        if not req.ok:
            raise SetupException(f"ERROR: Failed to create role: {name}")

        print("Role created")
    except SetupException as e:
        print(str(e))


def create_user(name, password, roles):
    try:
        print(f"\nCreating user {name}...")

        user = {"password": password,
                "roles": roles,
                "full_name": None,
                "email": None,
                "metadata": {}}

        req = requests.post(f"http://elastic:{ELASTIC_PASSWORD}@{ELASTIC_HOST}:9200/_security/user/{name}",
                            headers={"Content-Type": "application/json"},
                            data=json.dumps(user))
        if not req.ok:
            raise SetupException(f"ERROR: Failed to create user: {name}")

        print("User created")
    except SetupException as e:
        print(str(e))


def create_ilm_policy(name, rollover_age=None, rollover_size=None, rollover_doc_count=None,
                      warm_readonly=True, warm_age=None, delete_age=None):
    try:
        print(f"\nCreating ILM policy {name}...")

        data_base = {
            "policy": {
                "phases": {
                    "hot": {
                        "min_age": "0ms",
                        "actions": {
                            "set_priority": {
                                "priority": 100
                            }
                        }
                    },
                    "warm": {
                        "actions": {
                            "set_priority": {
                                "priority": 50
                            }
                        }
                    }
                }
            }
        }

        rollover_data = {}
        delete_data = {}

        if rollover_age:
            rollover_data['max_age'] = rollover_age
        if rollover_size:
            rollover_data['max_size'] = rollover_size
        if rollover_doc_count:
            rollover_data['max_docs'] = rollover_doc_count

        if delete_age:
            delete_data = {
                "min_age": delete_age,
                "actions": {
                    "delete": {}
                }
            }

        if rollover_data:
            data_base['policy']['phases']['hot']['actions']['rollover'] = rollover_data

        if delete_data:
            data_base['policy']['phases']['delete'] = delete_data

        if warm_readonly:
            data_base['policy']['phases']['warm']['actions']['readonly'] = {}

        if warm_age:
            data_base['policy']['phases']['warm']['min_age'] = warm_age

        pol_req = requests.put(f"http://elastic:{ELASTIC_PASSWORD}@{ELASTIC_HOST}:9200/_ilm/policy/{name}",
                               headers={"Content-Type": "application/json"},
                               data=json.dumps(data_base))
        if not pol_req.ok:
            raise SetupException(f"ERROR: Failed to create ILM policy: {name}")

        print("Policy created")
    except SetupException as e:
        print(str(e))


if __name__ == "__main__":
    ELASTIC_HOST = os.getenv("ELASTIC_HOST")
    ELASTIC_PASSWORD = os.getenv("ELASTIC_PASSWORD")
    APM_DELETE_AFTER = os.getenv("APM_DELETE_AFTER", "4d")
    FILEBEAT_DELETE_AFTER = os.getenv("FILEBEAT_DELETE_AFTER", "3d")
    METRICBEAT_DELETE_AFTER = os.getenv("METRICBEAT_DELETE_AFTER", "4d")

    # Testing connection to elasticsearch
    elastic_up = False
    while not elastic_up:
        try:
            req = requests.get(f"http://elastic:{ELASTIC_PASSWORD}@{ELASTIC_HOST}:9200/_cluster/health")
            if req.status_code == 401:
                print(f"Cannot configure elasticsearch server at {ELASTIC_HOST} with the provided password. Exiting...")
                if os.getenv("SU_USERNAME") and os.getenv("SU_PASSWORD"):
                    req = requests.get(f"http://{os.getenv('SU_USERNAME')}:{os.getenv('SU_PASSWORD')}"
                                       f"@{ELASTIC_HOST}:9200/_cluster/health")
                    if req.status_code == 401:
                        exit(1)
                exit(0)

            elastic_up = True
            print('Server is now online!')
        except ConnectionError:
            print(f"Waiting for elasticsearch server [{ELASTIC_HOST}] to come online...")
            time.sleep(1)

    # Setup APM ILM policies
    create_ilm_policy("apm-7.8.0-error", rollover_age="1d", rollover_size="5gb",
                      warm_readonly=True, delete_age=APM_DELETE_AFTER)
    create_ilm_policy("apm-7.8.0-metric", rollover_age="1d", rollover_size="5gb",
                      warm_readonly=True, delete_age=APM_DELETE_AFTER)
    create_ilm_policy("apm-7.8.0-span", rollover_age="1d", rollover_size="5gb",
                      warm_readonly=True, delete_age=APM_DELETE_AFTER)
    create_ilm_policy("apm-7.8.0-transaction", rollover_age="1d", rollover_size="5gb",
                      warm_readonly=True, delete_age=APM_DELETE_AFTER)

    # Setup filebeat ILM policies
    create_ilm_policy("filebeat-7.8.0", rollover_age="1d", rollover_size="20gb",
                      warm_readonly=True, delete_age=FILEBEAT_DELETE_AFTER)

    # Setup metricbeat ILM policies
    create_ilm_policy("metricbeat-7.8.0", rollover_age="1d", rollover_size="5gb",
                      warm_readonly=True, delete_age=METRICBEAT_DELETE_AFTER)

    # Creating kibana user
    if os.getenv("KIBANA_PASSWORD"):
        set_password("kibana", os.getenv("KIBANA_PASSWORD"))

    # Creating apm user
    if os.getenv("APM_PASSWORD"):
        create_role('apm', cluster_priv=["manage_ilm", "manage_pipeline"], indices=["apm-*", ".ml-anomalies*"])
        create_user('apm', os.getenv("APM_PASSWORD"), ["apm_user", "apm_system", "kibana_system", "apm"])

    # Creating filebeat user
    if os.getenv("FILEBEAT_PASSWORD"):
        create_role('filebeat', cluster_priv=["manage_ilm", "manage_pipeline"], indices=["filebeat-*"])
        create_user('filebeat', os.getenv("FILEBEAT_PASSWORD"), ["beats_system", "kibana_system", "filebeat"])

    # Creating metricbeat user
    if os.getenv("METRICBEAT_PASSWORD"):
        create_user('metricbeat', os.getenv("METRICBEAT_PASSWORD"), ["superuser"])

    # Creating assemblyline user
    if os.getenv("AL_PASSWORD"):
        create_role('assemblyline_system',
                    cluster_priv=["manage_index_templates", "manage_ilm"],
                    indices=[
                        "al_metrics_*",
                        "alert*",
                        "cached_*",
                        "emptyresult*",
                        "error*",
                        "file*",
                        "heuristic*",
                        "result*",
                        "service*",
                        "signature*",
                        "submission*",
                        "user*",
                        "vm*",
                        "workflow*"])
        create_user('assemblyline', os.getenv("AL_PASSWORD"), ["assemblyline_system"])

    # Creating new super user
    if os.getenv("SU_USERNAME"):
        if os.getenv("SU_PASSWORD"):
            create_user(os.getenv("SU_USERNAME"), os.getenv("SU_PASSWORD"), ["superuser"])

    # Creating Kibana only user
    if os.getenv("K_USERNAME"):
        if os.getenv("K_PASSWORD"):
            create_user(os.getenv("K_USERNAME"),
                        os.getenv("K_PASSWORD"),
                        ["kibana_user", "apm_user", "monitoring_user"])

    # Disable temporary password by setting a random one
    char_space = string.ascii_uppercase + string.digits + string.ascii_lowercase
    password = ''.join(random.SystemRandom().choice(char_space) for _ in range(32))
    set_password('elastic', password)



from assemblyline import odm
from . import PerformanceTimer

MSG_TYPES = {"WatcherHeartbeat"}
LOADER_CLASS = "assemblyline.odm.messages.watcher_heartbeat.WatcherMessage"


@odm.model()
class Metrics(odm.Model):
    expired = odm.Integer()
    cpu_seconds = PerformanceTimer()
    cpu_seconds_count = odm.Integer()
    busy_seconds = PerformanceTimer()
    busy_seconds_count = odm.Integer()


@odm.model()
class Heartbeat(odm.Model):
    instances = odm.Integer()
    metrics = odm.Compound(Metrics)
    watching = odm.Integer()


@odm.model()
class WatcherMessage(odm.Model):
    msg = odm.Compound(Heartbeat)
    msg_loader = odm.Enum(values={LOADER_CLASS}, default=LOADER_CLASS)
    msg_type = odm.Enum(values=MSG_TYPES, default="WatcherHeartbeat")
    sender = odm.Keyword()



# TODO: are we still using this?

"""
Dummy functions and values used to substitute for dynamic loaded methods that
have no interesting implementation by default.
"""


def always_false(*_, **__):
    return False


whitelist = []




import json
import logging
from urllib.parse import urlparse, parse_qs, unquote

import elasticapm

from assemblyline.common.exceptions import get_stacktrace_info
from assemblyline.filestore.transport.azure import TransportAzure
from assemblyline.filestore.transport.ftp import TransportFTP
from assemblyline.filestore.transport.http import TransportHTTP
from assemblyline.filestore.transport.local import TransportLocal
from assemblyline.filestore.transport.s3 import TransportS3
from assemblyline.filestore.transport.sftp import TransportSFTP


class FileStoreException(Exception):
    pass


class CorruptedFileStoreException(Exception):
    pass


def _get_extras(parsed_dict, valid_str_keys=None, valid_bool_keys=None):
    if not valid_str_keys:
        valid_str_keys = []
    if not valid_bool_keys:
        valid_bool_keys = []

    out = {}
    for k, v in parsed_dict.items():
        if k in valid_bool_keys:
            if v[0].lower() == 'true':
                out[k] = True
            elif v[0].lower() != 'true':
                out[k] = False
        if k in valid_str_keys:
            out[k] = v[0]

    return out


def create_transport(url, connection_attempts=None):
    """
    Transport are being initiated using an URL. They follow the normal url format:
    scheme://user:pass@host:port/path/to/file

    In this example, it will extract the following parameters:
    scheme: scheme
    host: host
    user: user
    password: pass
    port: port
    base: /path/to/file

    Certain transports can have extra parameters, those parameters need to be specified in the query part of the url.
    e.g.: sftp://host.com/path/to/file?private_key=/etc/ssl/pkey&private_key_pass=pass&validate_host=true
    scheme: sftp
    host: host.com
    user:
    password:
    private_key: /etc/ssl/pkey
    private_key_pass: pass
    validate_host: True

    NOTE: For transport with extra parameters, only specific extra parameters are allow. This is the list of extra
          parameter allowed:

          ftp: use_tls (bool)
          http: pki (string)
          sftp: private_key (string), private_key_pass (string), validate_host (bool)
          s3: aws_region (string), s3_bucket(string), use_ssl (bool), verify (bool)
          file: normalize (bool)

    """

    parsed = urlparse(url)

    base = parsed.path or '/'
    host = parsed.hostname
    if host == ".":
        base = "%s%s" % (host, base)
    port = parsed.port
    if parsed.password:
        password = unquote(parsed.password)
    else:
        password = ''
    user = parsed.username or ''

    scheme = parsed.scheme.lower()
    if scheme == 'ftp' or scheme == 'ftps':
        valid_bool_keys = ['use_tls']
        extras = _get_extras(parse_qs(parsed.query), valid_bool_keys=valid_bool_keys)
        if scheme == 'ftps':
            extras['use_tls'] = True

        t = TransportFTP(base=base, host=host, password=password, user=user, port=port, **extras)

    elif scheme == "sftp":
        valid_str_keys = ['private_key', 'private_key_pass']
        valid_bool_keys = ['validate_host']
        extras = _get_extras(parse_qs(parsed.query), valid_str_keys=valid_str_keys, valid_bool_keys=valid_bool_keys)

        t = TransportSFTP(base=base, host=host, password=password, user=user, **extras)

    elif scheme == 'http' or scheme == 'https':
        valid_str_keys = ['pki']
        valid_bool_keys = ['verify']
        extras = _get_extras(parse_qs(parsed.query), valid_str_keys=valid_str_keys, valid_bool_keys=valid_bool_keys)

        t = TransportHTTP(scheme=scheme, base=base, host=host, password=password, user=user, port=port, **extras)

    elif scheme == 'file':
        valid_bool_keys = ['normalize']
        extras = _get_extras(parse_qs(parsed.query), valid_bool_keys=valid_bool_keys)

        t = TransportLocal(base=base, **extras)

    elif scheme == 's3':
        valid_str_keys = ['aws_region', 's3_bucket']
        valid_bool_keys = ['use_ssl', 'verify']
        extras = _get_extras(parse_qs(parsed.query), valid_str_keys=valid_str_keys, valid_bool_keys=valid_bool_keys)

        t = TransportS3(base=base, host=host, port=port, accesskey=user, secretkey=password,
                        connection_attempts=connection_attempts, **extras)

    elif scheme == 'azure':
        valid_str_keys = ['access_key']
        extras = _get_extras(parse_qs(parsed.query), valid_str_keys=valid_str_keys)

        t = TransportAzure(base=base, host=host, connection_attempts=connection_attempts, **extras)

    else:
        raise FileStoreException("Unknown transport: %s" % scheme)

    return t


class FileStore(object):
    def __init__(self, *transport_urls, connection_attempts=None):
        self.log = logging.getLogger('assemblyline.transport')
        self.transports = [create_transport(url, connection_attempts) for url in transport_urls]
        self.local_transports = [
            t for t in self.transports if isinstance(t, TransportLocal)
        ]

    def __enter__(self):
        return self

    def __exit__(self, ex_type, exc_val, exc_tb):
        self.close()

    def __str__(self):
        return ', '.join(str(t) for t in self.transports)

    def close(self):
        for t in self.transports:
            try:
                t.close()
            except Exception as ex:
                trace = get_stacktrace_info(ex)
                self.log.warning('Transport problem: %s', trace)

    @elasticapm.capture_span(span_type='filestore')
    def delete(self, path, location='all'):
        with elasticapm.capture_span(name='delete', span_type='filestore', labels={'path': path}):
            for t in self.slice(location):
                try:
                    t.delete(path)
                except Exception as ex:
                    trace = get_stacktrace_info(ex)
                    self.log.info('Transport problem: %s', trace)

    @elasticapm.capture_span(span_type='filestore')
    def download(self, src_path, dest_path, location='any'):
        successful = False
        transports = []
        download_errors = []
        for t in self.slice(location):
            try:
                t.download(src_path, dest_path)
                transports.append(t)
                successful = True
                break
            except Exception as ex:
                download_errors.append((str(t), str(ex)))

        if not successful:
            raise FileStoreException('No transport succeeded => %s' % json.dumps(download_errors))
        return transports

    @elasticapm.capture_span(span_type='filestore')
    def exists(self, path, location='any'):
        transports = []
        for t in self.slice(location):
            try:
                if t.exists(path):
                    transports.append(t)
                    if location == 'any':
                        break
            except Exception as ex:
                trace = get_stacktrace_info(ex)
                self.log.warning('Transport problem: %s', trace)
        return transports

    @elasticapm.capture_span(span_type='filestore')
    def get(self, path, location='any'):
        for t in self.slice(location):
            try:
                if t.exists(path):
                    return t.get(path)
            except Exception as ex:
                trace = get_stacktrace_info(ex)
                self.log.warning('Transport problem: %s', trace)

    @elasticapm.capture_span(span_type='filestore')
    def put(self, dst_path, content, location='all', force=False):
        transports = []
        for t in self.slice(location):
            if force or not t.exists(dst_path):
                transports.append(t)
                t.put(dst_path, content)
                if not t.exists(dst_path):
                    raise FileStoreException('File transfer failed. Remote file does not '
                                             'exist for %s on %s (%s)' % (dst_path, location, t))
        return transports

    def slice(self, location):
        start, end = {
            'all': (0, len(self.transports)),
            'any': (0, len(self.transports)),
            'far': (-1, len(self.transports)),
            'near': (0, 1),
        }[location]

        transports = self.transports[start:end]
        assert(len(transports) >= 1)
        return transports

    @elasticapm.capture_span(span_type='filestore')
    def upload(self, src_path, dst_path, location='all', force=False):
        transports = []
        for t in self.slice(location):
            if force or not t.exists(dst_path):
                transports.append(t)
                t.upload(src_path, dst_path)
                if not t.exists(dst_path):
                    raise FileStoreException('File transfer failed. Remote file does not '
                                             'exist for %s on %s (%s)' % (dst_path, location, t))
        return transports

    @elasticapm.capture_span(span_type='filestore')
    def upload_batch(self, local_remote_tuples, location='all'):
        failed_tuples = []
        for (src_path, dst_path) in local_remote_tuples:
            try:
                self.upload(src_path, dst_path, location)
            except Exception as ex:
                trace = get_stacktrace_info(ex)
                failed_tuples.append((src_path, dst_path, trace))
        return failed_tuples



import hashlib
import json
import os
import random

from assemblyline.common import forge
from assemblyline.common.security import get_password_hash
from assemblyline.common.uid import get_random_id
from assemblyline.odm.models.alert import Alert
from assemblyline.odm.models.emptyresult import EmptyResult
from assemblyline.odm.models.error import Error
from assemblyline.odm.models.file import File
from assemblyline.odm.models.heuristic import Heuristic
from assemblyline.odm.models.result import Result
from assemblyline.odm.models.service import Service, UpdateSource
from assemblyline.odm.models.submission import Submission
from assemblyline.odm.models.user import User
from assemblyline.odm.models.user_settings import UserSettings
from assemblyline.odm.models.workflow import Workflow
from assemblyline.odm.randomizer import SERVICES, random_model_obj, get_random_phrase, get_random_uri, get_random_word
from assemblyline.run.suricata_importer import SuricataImporter
from assemblyline.run.yara_importer import YaraImporter
from assemblyline.datastore.helper import AssemblylineDatastore

full_file_list = []
classification = forge.get_classification()


class NullLogger(object):
    def info(self, msg):
        pass

    def warn(self, msg):
        pass

    def error(self, msg):
        pass

    def exception(self, msg):
        pass

    def warning(self, msg):
        pass


def create_alerts(ds, alert_count=50, submission_list=None, log=None):
    for _ in range(alert_count):
        a = random_model_obj(Alert)
        if isinstance(submission_list, list):
            submission = random.choice(submission_list)
            a.file.sha256 = submission.files[0].sha256
            a.sid = submission.sid

        a.owner = random.choice(['admin', 'user', 'other', None])
        ds.alert.save(a.alert_id, a)
        if log:
            log.info(f"\t{a.alert_id}")

    ds.alert.commit()


def create_heuristics(ds, log=None, heuristics_count=40):
    for srv in SERVICES.keys():
        for x in range(5):
            h = random_model_obj(Heuristic)
            h.heur_id = f"AL_{srv.upper()}_{x + 1}"
            h.name = get_random_phrase()
            ds.heuristic.save(h.heur_id, h)
            if log:
                log.info(f'\t{h.heur_id}')

    ds.heuristic.commit()


def create_services(ds: AssemblylineDatastore, log=None, limit=None):
    if not limit:
        limit = len(SERVICES)

    for svc_name, svc in list(SERVICES.items())[:limit]:
        service_data = {
            "name": svc_name,
            "enabled": True,
            "category": svc[0],
            "stage": svc[1],
            "version": "3.3.0",
            "docker_config": {
                "image": f"cccs/alsvc_{svc_name.lower()}:latest",
            },
        }

        if random.choice([True, False]):
            service_data['update_config'] = {
                "method": "run",
                "sources": [random_model_obj(UpdateSource)],
                "update_interval_seconds": 600,
                "generates_signatures": True
            }

        service_data = Service(service_data)
        # Save a v3 service
        ds.service.save(f"{service_data.name}_{service_data.version}", service_data)

        # Save the same service as v4
        service_data.version = "4.0.0"
        ds.service.save(f"{service_data.name}_{service_data.version}", service_data)

        # Save the default delta entry
        ds.service_delta.save(service_data.name, {"version": service_data.version})
        if log:
            log.info(f'\t{svc_name}')

    ds.service_delta.commit()
    ds.service.commit()


def create_signatures(ds):
    yara = YaraImporter(logger=NullLogger())
    suricata = SuricataImporter(logger=NullLogger())
    signatures = yara.import_file(get_yara_sig_path(), source="YAR_SAMPLE", default_status="DEPLOYED")
    signatures.extend(suricata.import_file(get_suricata_sig_path(), source="ET_SAMPLE", default_status="DEPLOYED"))

    ds.signature.commit()

    return [s['name'] for s in signatures]


def _create_errors_for_file(ds, f, services_done, log=None):
    e_list = []
    for _ in range(random.randint(0, 1)):
        e = random_model_obj(Error)

        # Only one error per service per file
        while e.response.service_name in services_done:
            e.response.service_name = random.choice(list(SERVICES.keys()))
        services_done.append(e.response.service_name)

        # Set the sha256
        e.sha256 = f

        e_key = e.build_key()
        e_list.append(e_key)
        if log:
            log.info(f"\t\t\t{e_key}")
        ds.error.save(e_key, e)

    return e_list


def _create_results_for_file(ds, f, possible_childs=None, log=None):
    r_list = []
    services_done = []
    section_depth_list = [[1, 1, 2, 3, 1], [1, 2, 1], [1, 2, 3, 1], [1, 2]]
    section_depth = random.choice(section_depth_list)
    for _ in range(random.randint(2, 5)):
        r = random_model_obj(Result)
        for depth_id, section in enumerate(r.result.sections):
            section.depth = section_depth[depth_id % len(section_depth)]
            if section.body_format == "GRAPH_DATA":
                cmap_min = 0
                cmap_max = random.choice([5, 10, 20])
                color_map_data = {
                    'type': 'colormap',
                    'data': {
                        'domain': [cmap_min, cmap_max],
                        'values': [random.random() * cmap_max for _ in range(50)]
                    }
                }
                section.body = json.dumps(color_map_data)
            elif section.body_format == "URL":
                data = [{"url": get_random_uri()} for _ in range(random.randint(1, 4))]
                section.body = json.dumps(data)
            elif section.body_format in ["JSON", "KEY_VALUE"]:
                data = {get_random_word(): get_random_id() for _ in range(random.randint(3, 9))}
                section.body = json.dumps(data)

        # Only one result per service per file
        while r.response.service_name in services_done:
            r.response.service_name = random.choice(list(SERVICES.keys()))
        services_done.append(r.response.service_name)

        # Set the sha256
        r.sha256 = f

        if random.randint(1, 10) > 8:
            # Generate and empty result
            r_key = f"{r.build_key()}.e"
            ds.emptyresult.save(r_key, random_model_obj(EmptyResult))
        else:
            r_key = r.build_key()
            # Set random extracted files that are not top level
            if not possible_childs:
                r.response.extracted = []
            else:
                for e in r.response.extracted:
                    e.sha256 = random.choice(possible_childs)

            # Set random supplementary files that are not top level
            if not possible_childs:
                r.response.supplementary = []
            else:
                for s in r.response.supplementary:
                    s.sha256 = random.choice(possible_childs)
            ds.result.save(r_key, r)

        if log:
            log.info(f"\t\t\t{r_key}")
        r_list.append(r_key)

    return r_list


def create_submission(ds, fs, log=None):
    f_list = []
    r_list = []
    e_list = []

    first_level_files = []
    s = random_model_obj(Submission)

    if log:
        log.info(f"\t{s.sid}")
        log.info(f"\tGenerating files for submission...")
    for _ in range(random.randint(4, 8)):
        f = random_model_obj(File)
        byte_str = get_random_phrase(wmin=8, wmax=20).encode()
        sha256 = hashlib.sha256(byte_str).hexdigest()
        f.sha256 = sha256
        ds.file.save(sha256, f)
        fs.put(sha256, byte_str)

        if log:
            log.info(f"\t\t{sha256}")

        f_list.append(sha256)

    for _ in range(random.randint(1, 2)):
        first_level_files.append(f_list.pop())

    if log:
        log.info(f"\t\tGenerating results and errors for top level files...")
    for f in first_level_files:
        r_list.extend(_create_results_for_file(ds, f, possible_childs=f_list, log=log))
        e_list.extend(_create_errors_for_file(ds, f, [x.split('.')[1] for x in r_list if x.startswith(f)], log=log))

    if log:
        log.info(f"\t\tGenerating results and errors for children files...")
    for f in f_list:
        r_list.extend(_create_results_for_file(ds, f, log=log))
        e_list.extend(_create_errors_for_file(ds, f, [x.split('.')[1] for x in r_list if x.startswith(f)], log=log))

    s.results = r_list
    s.errors = e_list

    s.error_count = len(e_list)
    s.file_count = len({x[:64] for x in r_list})

    s.files = s.files[:len(first_level_files)]

    fid = 0
    for f in s.files:
        f.sha256 = first_level_files[fid]
        fid += 1

    s.params.psid = None
    s.state = 'completed'

    ds.submission.save(s.sid, s)

    ds.emptyresult.commit()
    ds.error.commit()
    ds.file.commit()
    ds.result.commit()
    ds.submission.commit()

    return s


def create_users(ds, log=None):
    admin_pass = os.getenv("DEV_ADMIN_PASS", 'admin') or 'admin'
    user_pass = os.getenv("DEV_USER_PASS", 'user') or 'user'
    user_data = User({
        "agrees_with_tos": "NOW",
        "apikeys": {'devkey': {'acl': ["R", "W"], "password": get_password_hash(admin_pass)}},
        "classification": classification.RESTRICTED,
        "name": "Administrator",
        "email": "admin@assemblyline.local",
        "password": get_password_hash(admin_pass),
        "uname": "admin",
        "type": ["admin", "user", "signature_importer"]})
    ds.user.save('admin', user_data)
    ds.user_settings.save('admin', UserSettings({"ignore_cache": True, "deep_scan": True}))
    if log:
        log.info(f"\tU:{user_data.uname}   P:{admin_pass}")

    user_data = User({
        "name": "User",
        "email": "user@assemblyline.local",
        "apikeys": {'devkey': {'acl': ["R", "W"], "password": get_password_hash(user_pass)}},
        "password": get_password_hash(user_pass),
        "uname": "user"})
    ds.user.save('user', user_data)
    ds.user_settings.save('user', UserSettings())
    if log:
        log.info(f"\tU:{user_data.uname}   P:{user_pass}")

    ds.user.commit()


def create_workflows(ds, log=None):
    for _ in range(20):
        w_id = get_random_id()
        ds.workflow.save(w_id, random_model_obj(Workflow))
        if log:
            log.info(f'\t{w_id}')

    ds.workflow.commit()


def get_suricata_sig_path():
    for (d, _, filenames) in os.walk(__file__[:-11]):
        for f in filenames:
            if f == 'sample_suricata.rules':
                return os.path.join(d, f)

    raise Exception('Could not find suricata sample file...')


def get_yara_sig_path():
    for (d, _, filenames) in os.walk(__file__[:-11]):
        for f in filenames:
            if f == 'sample_rules.yar':
                return os.path.join(d, f)

    raise Exception('Could not find yara sample file...')


def wipe_alerts(ds):
    ds.alert.wipe()


def wipe_heuristics(ds):
    ds.heuristic.wipe()


def wipe_services(ds):
    ds.service.wipe()
    ds.service_delta.wipe()


def wipe_signatures(ds):
    ds.signature.wipe()


def wipe_submissions(ds, fs):
    ds.error.wipe()
    ds.file.wipe()
    ds.emptyresult.wipe()
    ds.result.wipe()
    ds.submission.wipe()
    ds.submission_summary.wipe()
    ds.submission_tree.wipe()

    for f in full_file_list:
        fs.delete(f)


def wipe_users(ds):
    ds.user.wipe()
    ds.user_settings.wipe()
    ds.user_avatar.wipe()
    ds.user_favorites.wipe()


def wipe_workflows(ds):
    ds.workflow.wipe()



from assemblyline import odm

MSG_TYPES = {"ScalerStatusHeartbeat"}
LOADER_CLASS = "assemblyline.odm.messages.scaler_status_heartbeat.ScalerStatusMessage"


@odm.model()
class Status(odm.Model):
    running = odm.Integer()
    target = odm.Integer()
    minimum = odm.Integer()
    maximum = odm.Integer()
    dynamic_maximum = odm.Integer()
    queue = odm.Integer()
    pressure = odm.Float()
    duty_cycle = odm.Float()


@odm.model()
class Heartbeat(odm.Model):
    service_name = odm.Keyword()
    metrics = odm.Compound(Status)


@odm.model()
class ScalerStatusMessage(odm.Model):
    msg = odm.Compound(Heartbeat)
    msg_loader = odm.Enum(values={LOADER_CLASS}, default=LOADER_CLASS)
    msg_type = odm.Enum(values=MSG_TYPES, default="ScalerStatusHeartbeat")
    sender = odm.Keyword()



import os
import platform
import re
import struct
import subprocess
import sys
import threading
import uuid
import zipfile
import msoffcrypto
from binascii import hexlify
from collections import defaultdict
from typing import Tuple, Union, Dict

import magic
import ssdeep
from cart import get_metadata_only

from assemblyline.common.digests import get_digests_for_file
from assemblyline.common.forge import get_constants
from assemblyline.common.str_utils import dotdump, safe_str

constants = get_constants()

STRONG_INDICATORS = {
    'code/vbs': [
        re.compile(rb'(^|\n)On Error Resume Next'),
        re.compile(rb'(^|\n)(?:Private)?[ \t]*Sub[ \t]+\w+\(*'),
        re.compile(rb'(^|\n)End Module'),
        re.compile(rb'(^|\n)ExecuteGlobal'),
        re.compile(rb'(^|\n)REM '),
        re.compile(rb'(ubound|lbound)\('),
    ],
    'code/javascript': [
        re.compile(rb'function([ \t]*|[ \t]+[\w]+[ \t]*)\([\w \t,]*\)[ \t]*{'),
        re.compile(rb'\beval[ \t]*\('),
        re.compile(rb'new[ \t]+ActiveXObject\('),
        re.compile(rb'xfa\.((resolve|create)Node|datasets|form)'),
        re.compile(rb'\.oneOfChild'),
        re.compile(rb'unescape\('),
        re.compile(rb'\.createElement\('),
    ],
    'code/csharp': [
        re.compile(rb'(^|\n)[ \t]*namespace[ \t]+[\w.]+'),
        re.compile(rb'(^|\n)[ \t]*using[ \t]+[\w.]+;'),
        re.compile(rb'(^|\n)[ \t]*internal class '),
    ],
    'code/php': [
        re.compile(rb'(^|\n)<\?php'),
        re.compile(rb'namespace[ \t]+[\w.]+'),
        re.compile(rb'function[ \t]*\w+[ \t]*\(\$[^)]+\)[ \t]*{'),
        re.compile(rb'\beval[ \t]*\('),
    ],
    'code/c': [
        re.compile(rb'(^|\n)(static|typedef)?[ \t]*struct '),
        re.compile(rb'(^|\n)#include[ \t]*([<"])[\w./]+([>"])'),
        re.compile(rb'(^|\n)#(ifndef |define |endif |pragma )'),
    ],
    'code/python': [
        re.compile(rb'(^|\n)[ \t]*if __name__[ \t]*==[ \t]*[\'\"]__main__[\'\"][ \t]*:'),
        re.compile(rb'(^|\n)[ \t]*from[ \t]+[\w.]+[ \t]+import[ \t]+[\w.*]+([ \t]+as \w+)?'),
        re.compile(rb'(^|\n)[ \t]*def[ \t]*\w+[ \t]*\([^)]*\)[ \t]*:'),
    ],
    'code/rust': [
        re.compile(rb'(^|\n)(pub|priv)[ \t]*(struct |enum |impl |const )'),
        re.compile(rb'(^|\n)[ \t]*fn[ \t]*\w+[ \t]*\(&self'),
        re.compile(rb'(println!|panic!)'),
    ],
    'code/lisp': [
        re.compile(rb'(^|\n)[ \t]*\((defmacro|defun|eval-when|in-package|list|export|defvar) '),
    ],
    'code/java': [
        re.compile(rb'(^|\n)[ \t]*public[ \t]+class[ \t]+\w+[ \t]+(extends[ \t]+\w+[ \t]+)?{'),
        re.compile(rb'(^|\n)[\w \t]+\([^)]+\)[ \t]+throws[ \t]+[\w, \t]+[ \t]+{'),
    ],
    'code/perl': [
        re.compile(rb'(^|\n)[ \t]*my[ \t]*\$\w+[ \t]*='),
        re.compile(rb'(^|\n)[ \t]*sub[ \t]*\w+[ \t]*{'),
    ],
    'code/ruby': [
        re.compile(rb'(^|\n)[ \t]*require(_all)?[ \t]*\'[\w/]+\''),
        re.compile(rb'rescue[ \t]+\w+[ \t]+=>'),
    ],
    'code/go': [
        re.compile(rb'(^|\n)[ \t]*import[ \t]+\('),
        re.compile(rb'(^|\n)[ \t]*func[ \t]+\w+\('),
    ],
    'code/css': [
        re.compile(rb'(^|\n|\})(html|body|footer|span\.|img\.|a\.|\.[a-zA-Z\-.]+)[^{]+{'
                   rb'[ \t]*(padding|color|width|margin|background|font|text)[^}]+\}'),
    ],
    'text/markdown': [
        re.compile(rb'\*[ \t]*`[^`]+`[ \t]*-[ \t]*\w+'),
    ],
    'document/email': [
        re.compile(rb'^Content-Type: ', re.MULTILINE),
        re.compile(rb'^Subject: ', re.MULTILINE),
        re.compile(rb'^MIME-Version: ', re.MULTILINE),
        re.compile(rb'^Message-ID: ', re.MULTILINE),
        re.compile(rb'^To: ', re.MULTILINE),
        re.compile(rb'^From: ', re.MULTILINE),
    ],
    'metadata/sysmon': [
        re.compile(rb'<Events>*'),
        re.compile(rb'<Event>*'),
        re.compile(rb'<\/Event>'),
        re.compile(rb'<\/Events>'),
    ],
    'code/xml': [
        # Check if it has an xml declaration header
        re.compile(rb'^\s*<\?xml[^>]+\?>', re.DOTALL | re.MULTILINE),
        # Check if it begins and ends with <tag ... and </tag ...> (for informal xml usages)
        re.compile(rb'^\s*<(?P<open>[\w:]+) .+</(?P=open)[^>]+>\s*$', re.DOTALL),
        # Check if a tag has an xmlns attribute
        re.compile(rb'<[^>]+xmlns[:=][^>]+>', re.MULTILINE),
    ],
    'code/ps1': [
        # Match one of the common Cmdlets
        re.compile(rb'(^|\n)(Get-ExecutionPolicy|Get-Service|Where-Object|ConvertTo-HTML|Select-Object|Get-Process|'
                   rb'Clear-History|ForEach-Object|Clear-Content|Compare-Object|New-ItemProperty|New-Object|'
                   rb'New-WebServiceProxy|Set-Alias|Wait-Job|Get-Counter|Test-Path|Get-WinEvent|Start-Sleep|'
                   rb'Set-Location|Get-ChildItem|Rename-Item|Stop-Process|Add-Type)'),
        # Match one of the common Classes
        re.compile(rb'(-memberDefinition|-Name|-namespace|-passthru)'),
        # Match one of the common Methods
        re.compile(rb'(\.Get(String|Field|Type|Method)\()')
    ]
}
STRONG_SCORE = 15
MINIMUM_GUESS_SCORE = 20

WEAK_INDICATORS = {
    'code/javascript': [rb'var ',
                        rb'document\.write\(',
                        rb'String\.(fromCharCode|raw)\(',
                        rb'Math\.(round|pow|sin|cos)\(',
                        rb'(isNaN|isFinite|parseInt|parseFloat)\(',
                        b'WSH',
                        ],
    'code/jscript': [rb'new[ \t]+ActiveXObject\(', rb'Scripting\.Dictionary'],
    'code/pdfjs': [rb'xfa\.((resolve|create)Node|datasets|form)', rb'\.oneOfChild'],
    'code/vbs': [
        rb'(^|\n)*[ ]{0,1000}[\t]*(Dim |Sub |Loop |Attribute |End Sub|Function |End Function )',
        b'CreateObject',
        b'WScript',
        b'window_onload',
        b'.SpawnInstance_',
        b'.Security_',
        b'WSH',
    ],
    'code/csharp': [rb'(^|\n)(protected)?[ \t]*override'],
    'code/sql': [rb'(^|\n)(create |drop |select |returns |declare )'],
    'code/php': [rb'\$this\->'],
    'code/c': [rb'(^|\n)(const char \w+;|extern |uint(8|16|32)_t )'],
    'code/python': [b'try:', b'except:', b'else:'],
    'code/java': [rb'(^|\n)[ \t]*package[ \t]+[\w\.]+;'],
    'code/perl': [rb'(^|\n)[ \t]*package[ \t]+[\w\.]+;', b'@_'],
    'text/markdown': [rb'\[[\w]+\]:[ \t]*http:'],
    'code/ps1': [
        # Check for PowerShell Parameters ex.  -Online -FeatureName
        rb'\s-([A-Z][a-z0-9]+)+',
        # Check for cmdlet names ex. Disable-WindowsOptionalFeature
        rb'([A-Z][a-z0-9]+)+-([A-Z][a-z0-9]+)+',
    ]
}
WEAK_SCORE = 1

WEAK_INDICATORS = {k: re.compile(b'|'.join(v)) for k, v in WEAK_INDICATORS.items()}

SHEBANG = re.compile(rb'^#![\w./]+/(?:env[ \t]*)?(\w+)[ \t]*\n')

EXECUTABLES = {
    'escript': 'erlang',
    'nush': 'nu',
    'macruby': 'ruby',
    'jruby': 'ruby',
    'rbx': 'ruby',
}

OLE_CLSID_GUIDs = {
    # GUID v0 (0)
    "00020803-0000-0000-C000-000000000046": "document/office/word",  # "MS Graph Chart"
    "00020900-0000-0000-C000-000000000046": "document/office/word",  # "MS Word95"
    "00020901-0000-0000-C000-000000000046": "document/office/word",  # "MS Word 6.0 - 7.0 Picture"
    "00020906-0000-0000-C000-000000000046": "document/office/word",  # "MS Word97"
    "00020907-0000-0000-C000-000000000046": "document/office/word",  # "MS Word"

    "00020C01-0000-0000-C000-000000000046": "document/office/excel",  # "Excel"
    "00020821-0000-0000-C000-000000000046": "document/office/excel",  # "Excel"
    "00020820-0000-0000-C000-000000000046": "document/office/excel",  # "Excel97"
    "00020810-0000-0000-C000-000000000046": "document/office/excel",  # "Excel95"

    "00021a14-0000-0000-C000-000000000046": "document/office/visio",  # "Visio"

    "0002CE02-0000-0000-C000-000000000046": "document/office/equation",  # "MS Equation 3.0"
    "0003000A-0000-0000-C000-000000000046": "document/office/paintbrush",  # "Paintbrush Picture",
    "0003000C-0000-0000-C000-000000000046": "document/office/package",  # "Package"
    "000C1084-0000-0000-C000-000000000046": "document/installer/windows",  # "Installer Package (MSI)"
    "00020D0B-0000-0000-C000-000000000046": "document/email",  # "MailMessage"

    # GUID v1 (Timestamp & MAC-48)
    "29130400-2EED-1069-BF5D-00DD011186B7": "document/office/wordpro",  # "Lotus WordPro"
    "46E31370-3F7A-11CE-BED6-00AA00611080": "document/office/word",  # "MS Forms 2.0 MultiPage"
    "5512D110-5CC6-11CF-8D67-00AA00BDCE1D": "document/office/word",  # "MS Forms 2.0 HTML SUBMIT"
    "5512D11A-5CC6-11CF-8D67-00AA00BDCE1D": "document/office/word",  # "MS Forms 2.0 HTML TEXT"
    "5512D11C-5CC6-11CF-8D67-00AA00BDCE1D": "document/office/word",  # "MS Forms 2.0 HTML Hidden"
    "64818D10-4F9B-11CF-86EA-00AA00B929E8": "document/office/powerpoint",  # "MS PowerPoint Presentation"
    "64818D11-4F9B-11CF-86EA-00AA00B929E8": "document/office/powerpoint",  # "MS PowerPoint Presentation"
    "11943940-36DE-11CF-953E-00C0A84029E9": "document/office/word",  # "MS Photo Editor 3.0 Photo"
    "B801CA65-A1FC-11D0-85AD-444553540000": "document/pdf",  # "Adobe Acrobat Document"
    "A25250C4-50C1-11D3-8EA3-0090271BECDD": "document/office/wordperfect",  # "WordPerfect Office"
    "C62A69F0-16DC-11CE-9E98-00AA00574A4F": "document/office/word",  # "Microsoft Forms 2.0 Form"
    "F4754C9B-64F5-4B40-8AF4-679732AC0607": "document/office/word",  # Word.Document.12
    "BDD1F04B-858B-11D1-B16A-00C0F0283628": "document/office/word",  # Doc (see CVE2012-0158)
}

recognized = constants.RECOGNIZED_TYPES

tag_to_extension = {
    'archive/chm': '.chm',
    'audiovisual/flash': '.swf',
    'code/batch': '.bat',
    'code/c': '.c',
    'code/csharp': '.cs',
    'code/hta': '.hta',
    'code/html': '.html',
    'code/java': '.java',
    'code/javascript': '.js',
    'code/jscript': '.js',
    'code/pdfjs': '.js',
    'code/perl': '.pl',
    'code/php': '.php',
    'code/ps1': '.ps1',
    'code/python': '.py',
    'code/ruby': '.rb',
    'code/vbs': '.vbs',
    'code/wsf': '.wsf',
    'document/installer/windows': '.msi',
    'document/office/excel': '.xls',
    'document/office/mhtml': '.mhtml',
    'document/office/ole': '.doc',
    'document/office/powerpoint': '.ppt',
    'document/office/rtf': '.doc',
    'document/office/unknown': '.doc',
    'document/office/visio': '.vsd',
    'document/office/word': '.doc',
    'document/office/wordperfect': 'wp',
    'document/office/wordpro': 'lwp',
    'document/pdf': '.pdf',
    'document/email': '.eml',
    'executable/windows/pe32': '.exe',
    'executable/windows/pe64': '.exe',
    'executable/windows/dll32': '.dll',
    'executable/windows/dll64': '.dll',
    'executable/windows/dos': '.exe',
    'executable/windows/com': '.exe',
    'executable/linux/elf32': '.elf',
    'executable/linux/elf64': '.elf',
    'executable/linux/so32': '.so',
    'executable/linux/so64': '.so',
    'java/jar': '.jar',
    'silverlight/xap': '.xap',
    'meta/shortcut/windows': '.lnk',
}

sl_patterns = [
    ['tnef', r'Transport Neutral Encapsulation Format'],
    ['chm', r'MS Windows HtmlHelp Data'],
    ['windows/dll64', r'^pe32\+ .*dll.*x86\-64'],
    ['windows/pe64', r'^pe32\+ .*x86\-64.*windows'],
    ['windows/dll32', r'^pe32 .*dll'],
    ['windows/pe32', r'^pe32 .*windows'],
    ['windows/pe', r'^pe unknown.*windows'],
    ['windows/dos', r'^(ms-)?dos executable'],
    ['windows/com', r'^com executable'],
    ['windows/dos', r'^8086 relocatable'],
    ['linux/elf32', r'^elf 32-bit lsb +executable'],
    ['linux/elf64', r'^elf 64-bit lsb +executable'],
    ['linux/so32', r'^elf 32-bit lsb +shared object'],
    ['linux/so64', r'^elf 64-bit lsb +shared object'],
    ['mach-o', r'^Mach-O'],
    ['7-zip', r'^7-zip archive data'],
    ['ace', r'^ACE archive data'],
    ['bzip2', r'^bzip2 compressed data'],
    ['cabinet', r'^installshield cab'],
    ['cabinet', r'^microsoft cabinet archive data'],
    ['cpio', r'cpio archive'],
    ['gzip', r'^gzip compressed data'],
    ['iso', r'ISO 9660'],
    ['lzma', r'^LZMA compressed data'],
    ['rar', r'^rar archive data'],
    ['tar', r'^(GNU|POSIX) tar archive'],
    ['ar', r'ar archive'],
    ['xz', r'^XZ compressed data'],
    ['zip', r'^zip archive data'],
    ['tcpdump', r'^tcpdump'],
    ['pdf', r'^pdf document'],
    ['bmp', r'^pc bitmap'],
    ['gif', r'^gif image data'],
    ['jpg', r'^jpeg image data'],
    ['png', r'^png image data'],
    ['installer/windows', r'(Installation Database|Windows Installer)'],
    ['office/excel', r'Microsoft.*Excel'],
    ['office/powerpoint', r'Microsoft.*PowerPoint'],
    ['office/word', r'Microsoft.*Word'],
    ['office/rtf', r'Rich Text Format'],
    ['office/ole', r'OLE 2'],
    ['office/unknown', r'Composite Document File|CDFV2'],
    ['office/unknown', r'Microsoft.*(OOXML|Document)'],
    ['office/unknown', r'Number of (Characters|Pages|Words)'],
    ['flash', r'Macromedia Flash'],
    ['autorun', r'microsoft windows autorun'],
    ['batch', r'dos batch file'],
    ['jar', r'[ (]Jar[) ]'],
    ['java', r'java program'],
    ['class', r'java class data'],
    ['perl', r'perl .*script'],
    ['php', r'php script'],
    ['python', r'python .*(script|byte)'],
    ['shell', r'(shell|sh) script'],
    ['xml', r'OpenGIS KML'],
    ['html', r'html'],
    ['sgml', r'sgml'],
    ['xml', r'xml'],
    ['sff', r'Frame Format'],
    ['shortcut/windows', r'^MS Windows shortcut'],
    ['email', r'Mime entity text'],
    ['sysmon', r'MS Windows Vista Event Log'],
]

sl_patterns = [[x[0], re.compile(x[1], re.IGNORECASE)] for x in sl_patterns]

sl_to_tl = {
    'windows/com': 'executable',
    'windows/dos': 'executable',
    'windows/pe32': 'executable',
    'windows/pe64': 'executable',
    'windows/dll32': 'executable',
    'windows/dll64': 'executable',
    'linux/elf32': 'executable',
    'linux/elf64': 'executable',
    'linux/so32': 'executable',
    'linux/so64': 'executable',
    'mach-o': 'executable',
    '7-zip': 'archive',
    'bzip2': 'archive',
    'cabinet': 'archive',
    'gzip': 'archive',
    'iso': 'archive',
    'rar': 'archive',
    'tar': 'archive',
    'zip': 'archive',
    'tcpdump': 'network',
    'pdf': 'document',
    'bmp': 'image',
    'gif': 'image',
    'jpg': 'image',
    'png': 'image',
    'shortcut/windows': 'meta',
}

# pylint:disable=C0301
tl_patterns = [
    ['document',
     r'Composite Document File|CDFV2|Corel|OLE 2|OpenDocument |Rich Text Format|Microsoft.*'
     r'(Document|Excel|PowerPoint|Word|OOXML)|Number of (Characters|Pages|Words)'],
    ['document', r'PostScript|pdf|MIME entity text'],
    ['java', r'jar |java'],
    ['code',
     r'Autorun|HTML |KML |LLVM |SGML |Visual C|XML |awk|batch |bytecode|perl|php|program|python'
     r'|ruby|scheme|script text exe|shell script|tcl'],
    ['network', r'capture'],
    ['unknown', r'CoreFoundation|Dreamcast|KEYBoard|OSF/Rose|Zope|quota|uImage'],
    ['unknown', r'disk|file[ ]*system|floppy|tape'],
    ['audiovisual',
     r'Macromedia Flash|Matroska|MIDI data|MPEG|MP4|MPG|MP3|QuickTime|RIFF|WebM|animation|audio|movie|music|ogg'
     r'|sound|tracker|video|voice data'],
    ['executable', r'803?86|COFF|ELF|Mach-O|ia32|executable|kernel|library|libtool|object'],
    ['unknown', r'Emulator'],
    ['image', r'DjVu|Surface|XCursor|bitmap|cursor|color|font|graphics|icon|image|jpeg'],
    ['archive',
     r'BinHex|InstallShield CAB|Transport Neutral Encapsulation Format|archive data|compress|mcrypt'
     r'|MS Windows HtmlHelp Data|current ar archive|cpio archive|ISO 9660'],
    ['meta', r'^MS Windows shortcut'],
    ['metadata', r'MS Windows Vista Event Log'],
    ['unknown', r'.*'],
]

trusted_mimes = {
    'application/x-bittorrent': 'meta/torrent',
    'application/x-tar': 'archive/tar',
    'message/rfc822': 'document/email',
    'text/calendar': 'text/calendar',
    'image/svg+xml': 'image/svg',
    'application/x-mach-binary': 'executable/mach-o',
    'application/vnd.ms-outlook': 'document/office/email',
    'application/x-iso9660-image': 'archive/iso',
}

tl_patterns = [[x[0], re.compile(x[1], re.IGNORECASE)] for x in tl_patterns]

custom = re.compile(r'^custom: ', re.IGNORECASE)

ssdeep_from_file = None

magic_lock = None
file_type = None
mime_type = None

if platform.system() != 'Windows':
    magic_lock = threading.Lock()

    file_type = magic.magic_open(magic.MAGIC_CONTINUE + magic.MAGIC_RAW)
    magic.magic_load(file_type, constants.RULE_PATH)

    mime_type = magic.magic_open(magic.MAGIC_CONTINUE + magic.MAGIC_RAW + magic.MAGIC_MIME)
    magic.magic_load(mime_type, constants.RULE_PATH)
    ssdeep_from_file = ssdeep.hash_from_file


# Translate the match object into a sub-type label.
def subtype(label: str) -> str:
    for entry in sl_patterns:
        if entry[1].search(label):  # pylint: disable=E1101
            return entry[0]

    return 'unknown'


def ident(buf, length: int, path) -> Dict:
    data = {'ascii': None, 'hex': None, 'magic': None, 'mime': None, 'type': 'unknown'}

    if length <= 0:
        return data

    header = buf[:min(64, length)]
    data['ascii'] = dotdump(header)
    data['hex'] = safe_str(hexlify(header))

    # noinspection PyBroadException
    try:
        # Loop over the labels returned by libmagic, ...
        labels = []
        if file_type:
            with magic_lock:
                labels = magic.magic_file(file_type, path).split(b'\n')
                labels = [label[2:].strip() if label.startswith(b'- ') else label.strip() for label in labels]

        mimes = []
        if mime_type:
            with magic_lock:
                mimes = magic.magic_file(mime_type, path).split(b'\n')
                mimes = [mime[2:].strip() if mime.startswith(b'- ') else mime.strip() for mime in mimes]

        # For user feedback set the mime and magic meta data to always be the primary
        # libmagic responses
        if len(labels) > 0:
            data['magic'] = safe_str(labels[0])

        if len(mimes) > 0 and mimes[0] != b'':
            data['mime'] = safe_str(mimes[0])

        # Highest priority is given to mime type matching something
        tagged = False

        for label in labels:
            label = dotdump(label)

            if custom.match(label):
                data['type'] = label.split('custom: ')[1].strip()
                tagged = True
                break

        # Second priority is mime times marked as trusted
        if not tagged:
            for mime in mimes:
                mime = dotdump(mime)

                if mime in trusted_mimes:
                    data['type'] = trusted_mimes[mime]
                    tagged = True
                    break

        # As a third priority try matching the tl_patterns
        if not tagged:
            minimum = len(tl_patterns)
            sl_tag = None

            # Try each label and see how far down the tl_patterns list we go
            # before we hit a match, the closer to the beginning of the list we are the better
            # the tag match is. The final line of tl_patterns matches anything and sets
            # tag to 'unknown', so this loop should never finish with sl_tag as None
            # Unless the tl_patters table has been changed inappropriately
            for label in labels:
                label = dotdump(label)

                # ... match against our patterns and, ...
                index = 0
                for entry in tl_patterns:
                    if index >= minimum:
                        break

                    if entry[1].search(label):  # pylint:disable=E1101
                        break

                    index += 1

                # ... keep highest precedence (lowest index) match.
                if index < minimum:
                    minimum = index
                    sl_tag = subtype(label)

                    # If a label does match, take the best from that label
                    # Further labels from magic are probably terrible
                    break

            assert sl_tag is not None, "tl_patterns seems to be missing a match all => unknown rule at the end"

            # Based on the sub tag we found, figure out the top level tag to use
            tl_tag = sl_to_tl.get(sl_tag, tl_patterns[minimum][0])
            data['type'] = '/'.join((tl_tag, sl_tag))

    except Exception as e:
        print(str(e))
        pass

    if not recognized.get(data['type'], False):
        data['type'] = 'unknown'

    if data['type'] == 'document/office/unknown':
        # noinspection PyBroadException
        try:
            root_entry_property_offset = buf.find(u"Root Entry".encode("utf-16-le"))
            if -1 != root_entry_property_offset:
                # Get root entry's GUID and try to guess document type
                clsid_offset = root_entry_property_offset + 0x50
                if len(buf) >= clsid_offset + 16:
                    clsid = buf[clsid_offset:clsid_offset + 16]
                    if len(clsid) == 16 and clsid != "\0" * len(clsid):
                        clsid_str = uuid.UUID(bytes_le=clsid)
                        clsid_str = clsid_str.urn.rsplit(':', 1)[-1].upper()
                        if clsid_str in OLE_CLSID_GUIDs:
                            data['type'] = OLE_CLSID_GUIDs[clsid_str]
        except Exception:
            pass

    return data


def _confidence(score: Union[int, float]) -> str:
    conf = float(score) / float(STRONG_SCORE * 5)
    conf = min(1.0, conf) * 100
    return str(int(conf)) + r'%'


def _differentiate(lang: str, scores_map: Dict) -> str:
    if lang == 'code/javascript':
        jscript_score = scores_map['code/jscript']
        pdfjs_score = scores_map['code/pdfjs']
        if pdfjs_score > 0 and pdfjs_score > jscript_score:
            return 'code/pdfjs'
        elif jscript_score > 0:
            return 'code/jscript'

    return lang


# Pass a filepath and this will return the guessed language in the AL tag format.
def guess_language(path: str) -> Tuple[str, Union[str, int]]:
    file_length = os.path.getsize(path)
    with open(path, 'rb') as fh:
        if file_length > 131070:
            buf = fh.read(65535)
            fh.seek(file_length - 65535)
            buf += fh.read(65535)
        else:
            buf = fh.read()

    scores = defaultdict(int)
    shebang_lang = re.match(SHEBANG, buf)
    if shebang_lang:
        lang = shebang_lang.group(1)
        lang = 'code/' + EXECUTABLES.get(safe_str(lang), safe_str(lang))
        scores[lang] = STRONG_SCORE * 3

    for lang, patterns in STRONG_INDICATORS.items():
        for pattern in patterns:
            for _ in re.findall(pattern, buf):
                scores[lang] += STRONG_SCORE

    for lang, pattern in WEAK_INDICATORS.items():
        for _ in re.findall(pattern, buf):
            scores[lang] += WEAK_SCORE

    for lang in list(scores.keys()):
        if scores[lang] < MINIMUM_GUESS_SCORE:
            scores.pop(lang)

    max_v = 0
    if len(scores) > 0:
        max_v = max(list(scores.values()))
    high_scores = [(k, v) for k, v in scores.items() if v == max_v]
    high_scores = [(_differentiate(k, scores), v) for k, v in high_scores]

    if len(high_scores) != 1:
        return 'unknown', 0
    else:
        confidences = [(k, _confidence(v)) for k, v in high_scores]
        return confidences[0]


# noinspection PyBroadException
def zip_ident(path: str, fallback: str = None) -> str:
    file_list = []

    try:
        with zipfile.ZipFile(path, "r") as zf:
            file_list = [zfname for zfname in zf.namelist()]
    except Exception:
        try:
            stdout, _ = subprocess.Popen(["unzip", "-l", path],
                                         stderr=subprocess.PIPE,
                                         stdout=subprocess.PIPE).communicate()
            lines = stdout.splitlines()
            index = lines[1].index(b"Name")
            for file_name in lines[3:-2]:
                file_list.append(safe_str(file_name[index:]))
        except Exception:
            if fallback is not None:
                return fallback

    tot_files = 0
    tot_class = 0
    tot_jar = 0

    is_ipa = False
    is_jar = False
    is_word = False
    is_excel = False
    is_ppt = False
    doc_props = False
    doc_rels = False
    doc_types = False
    android_manifest = False
    android_dex = False

    for file_name in file_list:
        if file_name[:8] == 'META-INF' and file_name[9:] == 'MANIFEST.MF':
            is_jar = True
        elif file_name == 'AndroidManifest.xml':
            android_manifest = True
        elif file_name == 'classes.dex':
            android_dex = True
        elif file_name.startswith('Payload/') and file_name.endswith(".app/Info.plist"):
            is_ipa = True
        elif file_name.endswith(".class"):
            tot_class += 1
        elif file_name.endswith(".jar"):
            tot_jar += 1
        elif file_name.startswith('word/'):
            is_word = True
        elif file_name.startswith('xl/'):
            is_excel = True
        elif file_name.startswith('ppt/'):
            is_ppt = True
        elif file_name.startswith('docProps/'):
            doc_props = True
        elif file_name.startswith('_rels/'):
            doc_rels = True
        elif file_name == '[Content_Types].xml':
            doc_types = True

        tot_files += 1

    if 0 < tot_files < (tot_class + tot_jar) * 2:
        is_jar = True

    if is_jar and android_manifest and android_dex:
        return 'android/apk'
    elif is_ipa:
        return 'ios/ipa'
    elif is_jar:
        return 'java/jar'
    elif (doc_props or doc_rels) and doc_types:
        if is_word:
            return 'document/office/word'
        elif is_excel:
            return 'document/office/excel'
        elif is_ppt:
            return 'document/office/powerpoint'
        else:
            return 'document/office/unknown'
    else:
        return 'archive/zip'


# noinspection PyBroadException
def cart_ident(path: str) -> str:
    try:
        metadata = get_metadata_only(path)
    except Exception:
        return 'corrupted/cart'
    return metadata.get('al', {}).get('type', 'archive/cart')


def dos_ident(path: str) -> str:
    # noinspection PyBroadException
    try:
        with open(path, "rb") as fh:
            file_header = fh.read(0x40)
            if file_header[0:2] != "MZ":
                raise ValueError()

            header_pos, = struct.unpack("<I", file_header[-4:])
            fh.seek(header_pos)
            if fh.read(4) != "PE\x00\x00":
                raise ValueError()
            machine_id, = struct.unpack("<H", fh.read(2))
            if machine_id == 0x014c:
                width = 32
            elif machine_id == 0x8664:
                width = 64
            else:
                raise ValueError()
            characteristics, = struct.unpack("<H", fh.read(18)[-2:])
            if characteristics & 0x2000:
                pe_type = "dll"
            elif characteristics & 0x0002:
                pe_type = "pe"
            else:
                raise ValueError()
            return "executable/windows/%s%i" % (pe_type, width)
    except Exception:
        pass
    return "executable/windows/dos"


def fileinfo(path: str) -> Dict:
    path = safe_str(path)

    data = get_digests_for_file(path, on_first_block=ident)

    # This is a special case, we know if the mime is set to one of these values
    # then the input file is almost certainly an office file, but based on only the first
    # block magic can't figure out any more than that. To handle that case we will read the
    # entire file, and identify again.
    if data['mime'] is not None and data['mime'].lower() in ['application/cdfv2-corrupt', 'application/cdfv2-unknown']:
        with open(path, 'rb') as fh:
            buf = fh.read()
            buflen = len(buf)
            data.update(ident(buf, buflen))
    data['ssdeep'] = ssdeep_from_file(path) if ssdeep_from_file else ''

    # When data is parsed from a cart file we trust its metatdata and can skip the recognition test later
    cart_metadata_set = False

    if not int(data.get('size', -1)):
        data['type'] = 'empty'
    elif data['type'] in ['archive/zip', 'java/jar']:
        # In addition to explicit zip files, we also want to run zip_ident when
        # a file is a jar as there is a high rate of false positive (magic
        # matching eclipse and other java related files as jars)
        data['type'] = zip_ident(path)
    elif data['type'] == 'document/office/unknown':
        # For unknown document files try identifying them by unziping,
        # but don't commit to it being a zip if it can't be extracted
        data['type'] = zip_ident(path, data['type'])
    elif data['type'] == 'unknown':
        data['type'], _ = guess_language(path)
    elif data['type'] == 'archive/cart':
        data['type'] = cart_ident(path)
        cart_metadata_set = True
    elif data['type'] == 'executable/windows/dos':
        # The default magic file misidentifies PE files with a munged DOS header
        data['type'] = dos_ident(path)
    elif data['type'] == 'code/html':
        # Magic detects .hta files as .html, guess_language detects .hta files as .js/.vbs
        # If both conditions are met, it's fair to say that the file is an .hta
        lang, _ = guess_language(path)
        if lang in ["code/javascript", "code/vbs"]:
            data['type'] = 'code/hta'

    if data['type'] in ['document/office/word', 'document/office/excel',
                        'document/office/powerpoint', 'document/office/unknown']:
        try:
            msoffcrypto_obj = msoffcrypto.OfficeFile(open(path, "rb"))
            if msoffcrypto_obj and msoffcrypto_obj.is_encrypted():
                data['type'] = 'document/office/passwordprotected'
        except Exception:
            # If msoffcrypto can't handle the file to confirm that it is/isn't password protected,
            # then it's not meant to be. Moving on!
            pass

    if not recognized.get(data['type'], False) and not cart_metadata_set:
        data['type'] = 'unknown'

    return data


if __name__ == '__main__':
    from pprint import pprint

    # noinspection PyBroadException
    if len(sys.argv) > 1:
        pprint(fileinfo(sys.argv[1]))
    else:
        name = sys.stdin.readline().strip()
        while name:
            a = fileinfo(name)
            print('\t'.join(dotdump(str(a[k])) for k in ('type', 'ascii',
                                                         'entropy', 'hex', 'magic',
                                                         'mime', 'md5', 'sha1', 'sha256',
                                                         'ssdeep', 'size')))
            name = sys.stdin.readline().strip()



from assemblyline import odm


@odm.model(index=False, store=False)
class EnvironmentVariable(odm.Model):
    name = odm.Keyword()
    value = odm.Keyword()


@odm.model(index=False, store=False)
class DockerConfigDelta(odm.Model):
    allow_internet_access = odm.Optional(odm.Boolean())
    command = odm.Optional(odm.List(odm.Keyword()))
    cpu_cores = odm.Optional(odm.Float())
    environment = odm.Optional(odm.List(odm.Compound(EnvironmentVariable)))
    image = odm.Optional(odm.Keyword())  # The docker image and tag, optionally including registry in the normal way
    registry_username = odm.Optional(odm.Keyword())  # The username to use when pulling the image
    registry_password = odm.Optional(odm.Keyword())  # The password or token to use when pulling the image
    ports = odm.Optional(odm.List(odm.Keyword()))
    ram_mb = odm.Optional(odm.Integer())
    ram_mb_min = odm.Optional(odm.Integer())


@odm.model(index=False, store=False)
class UpdateSourceDelta(odm.Model):
    headers = odm.Optional(odm.List(odm.Compound(EnvironmentVariable)))
    name = odm.Optional(odm.Keyword())
    password = odm.Optional(odm.Keyword(default=""))
    pattern = odm.Optional(odm.Keyword(default=""))
    private_key = odm.Optional(odm.Keyword(default=""))
    ca_cert = odm.Optional(odm.Keyword(default=""))
    ssl_ignore_errors = odm.Boolean(default=False)
    proxy = odm.Optional(odm.Keyword(default=""))
    uri = odm.Optional(odm.Keyword())
    username = odm.Optional(odm.Keyword(default=""))
    default_classification = odm.Optional(odm.Classification())


@odm.model(index=False, store=False)
class PersistentVolumeDelta(odm.Model):
    mount_path = odm.Optional(odm.Keyword())
    capacity = odm.Optional(odm.Keyword())
    storage_class = odm.Optional(odm.Keyword())


@odm.model(index=False, store=False)
class DependencyConfigDelta(odm.Model):
    container = odm.Optional(odm.Compound(DockerConfigDelta))
    volumes = odm.Mapping(odm.Compound(PersistentVolumeDelta), default={})


@odm.model(index=False, store=False)
class UpdateConfigDelta(odm.Model):
    # build_options = odm.Optional(odm.Compound(DockerfileConfigDelta))
    generates_signatures = odm.Optional(odm.Boolean(), index=True)
    method = odm.Optional(odm.Enum(values=['run', 'build']))
    run_options = odm.Optional(odm.Compound(DockerConfigDelta))
    sources = odm.Optional(odm.List(odm.Compound(UpdateSourceDelta)))
    update_interval_seconds = odm.Optional(odm.Integer())
    wait_for_update = odm.Optional(odm.Boolean())


@odm.model(index=False, store=False)
class SubmissionParamsDelta(odm.Model):
    default = odm.Optional(odm.Any())
    name = odm.Optional(odm.Keyword())
    type = odm.Optional(odm.Enum(values=['str', 'int', 'list', 'bool']))
    value = odm.Optional(odm.Any())
    list = odm.Optional(odm.Any())


@odm.model(index=True, store=False)
class ServiceDelta(odm.Model):
    accepts = odm.Optional(odm.Keyword(), store=True)
    rejects = odm.Optional(odm.Keyword(), store=True)

    category = odm.Optional(odm.Keyword(), store=True, copyto="__text__")
    config = odm.Optional(odm.Mapping(odm.Any()), index=False)
    description = odm.Optional(odm.Text(), store=True, copyto="__text__")
    default_result_classification = odm.Optional(odm.ClassificationString())
    enabled = odm.Optional(odm.Boolean(), store=True)
    is_external = odm.Optional(odm.Boolean())
    licence_count = odm.Optional(odm.Integer())

    name = odm.Optional(odm.Keyword(), store=True, copyto="__text__")
    version = odm.Keyword(store=True)

    disable_cache = odm.Optional(odm.Boolean())

    stage = odm.Optional(odm.Keyword(), store=True, copyto="__text__")
    submission_params = odm.Optional(odm.List(odm.Compound(SubmissionParamsDelta)), index=False)
    timeout = odm.Optional(odm.Integer())

    docker_config: DockerConfigDelta = odm.Optional(odm.Compound(DockerConfigDelta))
    dependencies: DependencyConfigDelta = odm.Mapping(odm.Compound(DependencyConfigDelta), default={})

    update_channel = odm.Optional(odm.Enum(values=["stable", "rc", "beta", "dev"]))
    update_config: UpdateConfigDelta = odm.Optional(odm.Compound(UpdateConfigDelta))



import functools

from assemblyline.remote.datatypes.queues.priority import PriorityQueue


def determine_dispatcher(sid, shards):
    n = functools.reduce(lambda x, y: x ^ y, [int(y, 16) for y in sid[-12:]])
    return n % shards


class DispatchQueue(object):
    def __init__(self, host=None, port=None, shards=None):
        self.host = host or '127.0.0.1'
        self.port = int(port or 6379)
        self.shards = int(shards or 1)

        self.queues = {}

    def _get_queue(self, name):
        q = self.queues.get(name, None)
        if not q:
            self.queues[name] = q = PriorityQueue(name, self.host, self.port)
        return q

    def length(self, name):
        return self._get_queue(name).length()

    def pop(self, name, num=1):
        return self._get_queue(name).pop(num)

    def send(self, message_id, message, shards, priority, dispatch_queue=None):
        if priority is None:
            priority = 0

        n = determine_dispatcher(message_id, shards)
        if not dispatch_queue:
            dispatch_queue = 'ingest-queue-' + str(n)
        self._get_queue(dispatch_queue).push(priority, message)



from collections.abc import Mapping
from typing import Dict, Optional, Mapping as _Mapping, Union


def recursive_update(d: Dict, u: _Mapping) -> Union[Dict, _Mapping]:
    if d is None:
        return u

    if u is None:
        return d

    for k, v in u.items():
        if isinstance(v, Mapping):
            d[k] = recursive_update(d.get(k, {}), v)
        else:
            d[k] = v

    return d


def get_recursive_delta(d1: (Dict, Mapping), d2: (Dict, Mapping)) -> Dict:
    if d1 is None:
        return d2

    if d2 is None:
        return d1

    out = {}
    for k1, v1 in d1.items():
        if isinstance(v1, Mapping):
            internal = get_recursive_delta(v1, d2.get(k1, {}))
            if internal:
                out[k1] = internal
        else:
            if k1 in d2:
                v2 = d2[k1]
                if v1 != v2:
                    out[k1] = v2

    for k2, v2 in d2.items():
        if k2 not in d1:
            out[k2] = v2

    return out


def flatten(data: Dict, parent_key: Optional[str] = None) -> Dict:
    items = []
    for k, v in data.items():
        cur_key = f"{parent_key}.{k}" if parent_key is not None else k
        if isinstance(v, dict):
            items.extend(flatten(v, cur_key).items())
        else:
            items.append((cur_key, v))

    return dict(items)


def unflatten(data: Dict) -> Dict:
    out = dict()
    for k, v in data.items():
        parts = k.split(".")
        d = out
        for p in parts[:-1]:
            if p not in d:
                d[p] = dict()
            d = d[p]
        d[parts[-1]] = v
    return out



import logging

from assemblyline.common.attack_map import attack_map, software_map, group_map, revoke_map

heur_logger = logging.getLogger("assemblyline.heuristics")


def service_heuristic_to_result_heuristic(srv_heuristic, heuristics):
    heur_id = srv_heuristic['heur_id']
    attack_ids = srv_heuristic.pop('attack_ids', [])
    signatures = srv_heuristic.pop('signatures', {})
    frequency = srv_heuristic.pop('frequency', 0)
    score_map = srv_heuristic.pop('score_map', {})

    # Validate the heuristic and recalculate its score
    heuristic = Heuristic(heur_id, attack_ids, signatures, score_map, frequency, heuristics)

    try:
        # Assign the newly computed heuristic to the section
        output = dict(
            heur_id=heur_id,
            score=heuristic.score,
            name=heuristic.name,
            attack=[],
            signature=[]
        )

        # Assign the multiple attack IDs to the heuristic
        for attack_id in heuristic.attack_ids:
            attack_item = dict(
                attack_id=attack_id,
                pattern=attack_map[attack_id]['name'],
                categories=attack_map[attack_id]['categories']
            )
            output['attack'].append(attack_item)

        # Assign the multiple signatures to the heuristic
        for sig_name, freq in heuristic.signatures.items():
            signature_item = dict(
                name=sig_name,
                frequency=freq
            )
            output['signature'].append(signature_item)

        return output, heuristic.associated_tags
    except InvalidHeuristicException as e:
        heur_logger.warning(str(e))
        raise


class InvalidHeuristicException(Exception):
    pass


class Heuristic(object):
    def __init__(self, heur_id, attack_ids, signatures, score_map, frequency, heuristics):
        # Validate heuristic
        definition = heuristics.get(heur_id)
        if not definition:
            raise InvalidHeuristicException(f"Heuristic with ID '{heur_id}' does not exist, skipping...")

        # Set defaults
        self.heur_id = heur_id
        self.attack_ids = []
        self.name = definition.name
        self.classification = definition.classification
        self.associated_tags = []

        # Show only attack_ids that are valid
        attack_ids = attack_ids or []
        attack_ids = [revoke_map.get(x, x) for x in attack_ids]
        for a_id in attack_ids:
            if a_id in attack_map:
                self.attack_ids.append(a_id)
            elif a_id in software_map:
                software_def = software_map[a_id]
                implant_name = software_def.get('name', None)
                if implant_name:
                    self.associated_tags.append(('attribution.implant', implant_name))

                for s_a_id in software_def['attack_ids']:
                    if s_a_id in attack_map:
                        self.attack_ids.append(s_a_id)
                    else:
                        heur_logger.warning(f"Invalid related attack_id '{s_a_id}' for software '{a_id}' "
                                            f"in heuristic '{heur_id}'. Ignoring it.")
            elif a_id in group_map:
                group_name = group_map[a_id].get('name', None)
                if group_name:
                    self.associated_tags.append(('attribution.actor', group_name))
            else:
                heur_logger.warning(f"Invalid attack_id '{a_id}' in heuristic '{heur_id}'. Ignoring it.")
        self.attack_ids = list(set(self.attack_ids))

        # Calculate the score for the signatures
        self.signatures = signatures or {}
        if len(self.signatures) > 0:
            self.score = 0
            for sig_name, freq in signatures.items():
                sig_score = definition.signature_score_map.get(sig_name, score_map.get(sig_name, definition.score))
                self.score += sig_score * freq
        else:
            # Calculate the score for the heuristic frequency
            frequency = frequency or 1
            self.score = definition.score * frequency

        # Check scoring boundaries
        if definition.max_score:
            self.score = min(self.score, definition.max_score)



from assemblyline.odm import Keyword, Text, List, Compound, Date, Integer, \
    Float, Boolean, Mapping, Classification, Enum, Any, UUID, Optional, IP, Domain, URI, URIPath, MAC, PhoneNumber, \
    SSDeepHash, SHA1, SHA256, MD5, Platform, Processor, ClassificationString, FlattenedObject, Email

# Simple types can be resolved by a direct mapping
__type_mapping = {
    Keyword: 'keyword',
    Boolean: 'boolean',
    Integer: 'integer',
    Float: 'float',
    Date: 'date',
    Text: 'text',
    Classification: 'keyword',
    ClassificationString: 'keyword',
    Enum: 'keyword',
    UUID: 'keyword',
    IP: 'ip',
    Domain: 'keyword',
    Email: 'keyword',
    URI: 'keyword',
    URIPath: 'keyword',
    MAC: 'keyword',
    PhoneNumber: 'keyword',
    SSDeepHash: 'text',
    SHA1: 'keyword',
    SHA256: 'keyword',
    MD5: 'keyword',
    Platform: 'keyword',
    Processor: 'keyword',
    FlattenedObject: 'nested'
}
__analyzer_mapping = {
    SSDeepHash: 'text_fuzzy',
}
__normalizer_mapping = {
    SHA1: 'lowercase_normalizer',
    SHA256: 'lowercase_normalizer',
    MD5: 'lowercase_normalizer',
}
# TODO: We might want to use custom analyzers for Classification and Enum and not create special backmapping cases
back_mapping = {v: k for k, v in __type_mapping.items() if k not in [Enum, Classification, UUID, IP, Domain, URI,
                                                                     URIPath, MAC, PhoneNumber, SSDeepHash, Email,
                                                                     SHA1, SHA256, MD5, Platform, Processor,
                                                                     ClassificationString]}
back_mapping.update({x: Keyword for x in set(__analyzer_mapping.values())})


def build_mapping(field_data, prefix=None, allow_refuse_implicit=True):
    """
    The mapping for Elasticsearch based on a python model object.
    """

    prefix = prefix or []
    mappings = {}
    dynamic = []

    def set_mapping(temp_field, body):
        body['index'] = temp_field.index
        body['store'] = temp_field.store
        if temp_field.copyto:
            assert len(temp_field.copyto) == 1
            body['copy_to'] = temp_field.copyto[0]

        return body

    # Fill in the sections
    for field in field_data:
        path = prefix + ([field.name] if field.name else [])
        name = '.'.join(path)

        if isinstance(field, Classification):
            mappings[name.strip(".")] = set_mapping(field, {
                'type': __type_mapping[field.__class__]
            })
            if "." not in name:
                mappings.update({
                    "__access_lvl__": {
                        'type': 'integer',
                        'store': False,
                        'index': True},
                    "__access_req__": {
                        'type': 'keyword',
                        'store': False,
                        'index': True},
                    "__access_grp1__": {
                        'type': 'keyword',
                        'store': False,
                        'index': True},
                    "__access_grp2__": {
                        'type': 'keyword',
                        'store': False,
                        'index': True},
                })

        elif isinstance(field, (Boolean, Integer, Float, Text)):
            mappings[name.strip(".")] = set_mapping(field, {
                'type': __type_mapping[field.__class__]
            })

        elif field.__class__ in __analyzer_mapping:
            mappings[name.strip(".")] = set_mapping(field, {
                'type': __type_mapping[field.__class__],
                "analyzer": __analyzer_mapping[field.__class__]
            })

        elif isinstance(field, Keyword):
            es_data_type = __type_mapping[field.__class__]
            data = {'type': es_data_type}
            if es_data_type == "keyword":
                data["ignore_above"] = 8191  # The maximum always safe value in elasticsearch
            if field.__class__ in __normalizer_mapping:
                data['normalizer'] = __normalizer_mapping[field.__class__]
            mappings[name.strip(".")] = set_mapping(field, data)

        elif isinstance(field, Date):
            mappings[name.strip(".")] = set_mapping(field, {
                'type': __type_mapping[field.__class__],
                'format': 'date_optional_time||epoch_millis',
            })

        elif isinstance(field, FlattenedObject):
            if not field.index or isinstance(field.child_type, Any):
                mappings[name.strip(".")] = {"type": "object", "enabled": False}
            else:
                dynamic.extend(build_templates(f'{name}.*', field.child_type, nested_template=True, index=field.index))

        elif isinstance(field, List):
            temp_mappings, temp_dynamic = build_mapping([field.child_type], prefix=path,
                                                        allow_refuse_implicit=False)
            mappings.update(temp_mappings)
            dynamic.extend(temp_dynamic)

        elif isinstance(field, Optional):
            temp_mappings, temp_dynamic = build_mapping([field.child_type], prefix=prefix,
                                                        allow_refuse_implicit=False)
            mappings.update(temp_mappings)
            dynamic.extend(temp_dynamic)

        elif isinstance(field, Compound):
            temp_mappings, temp_dynamic = build_mapping(field.fields().values(), prefix=path,
                                                        allow_refuse_implicit=False)
            mappings.update(temp_mappings)
            dynamic.extend(temp_dynamic)

        elif isinstance(field, Mapping):
            if not field.index or isinstance(field.child_type, Any):
                mappings[name.strip(".")] = {"type": "object", "enabled": False}
            else:
                dynamic.extend(build_templates(f'{name}.*', field.child_type, index=field.index))

        elif isinstance(field, Any):
            field_template = {
                "path_match": name,
                "mapping": {
                    "type": "keyword",
                    "index": False,
                    "store": False
                }
            }

            if field.index or field.store:
                raise ValueError(f"Any may not be indexed or stored: {name}")
            dynamic.append({f"{name}_tpl": field_template})

        else:
            raise NotImplementedError(f"Unknown type for elasticsearch schema: {field.__class__}")

    # The final template must match everything and disable indexing
    # this effectively disables dynamic indexing EXCEPT for the templates
    # we have defined
    if not dynamic and allow_refuse_implicit:
        # We cannot use the dynamic type matching if others are in play because they conflict with each other
        # TODO: Find a way to make them work together.
        dynamic.append({'refuse_all_implicit_mappings': {
            "match": "*",
            "mapping": {
                "index": False,
                "ignore_malformed": True,
            }
        }})

    return mappings, dynamic


def build_templates(name, field, nested_template=False, index=True) -> list:
    if isinstance(field, (Keyword, Boolean, Integer, Float, Text)):
        if nested_template:
            main_template = {
                "match": f"{name}",
                "mapping": {
                    "type": "nested"
                }
            }

            return [{f"nested_{name}": main_template}]
        else:
            field_template = {
                "path_match": name,
                "mapping": {
                    "type": __type_mapping[field.__class__],
                }
            }

            field_template['mapping']['index'] = field.index
            field_template['mapping']['store'] = field.store
            if field.copyto:
                assert len(field.copyto) == 1
                field_template['mapping']['copy_to'] = field.copyto[0]

            return [{f"{name}_tpl": field_template}]

    elif isinstance(field, Any) or not index:
        field_template = {
            "path_match": name,
            "mapping": {
                "type": "keyword",
                "index": False,
                "store": False
            }
        }

        if field.index or field.store:
            raise ValueError(f"Mapping to Any may not be indexed or stored: {name}")
        return [{f"{name}_tpl": field_template}]

    elif isinstance(field, (Mapping, List)):
        temp_name = name
        if field.name:
            temp_name = f"{name}.{field.name}"
        return build_templates(temp_name, field.child_type, nested_template=True)

    elif isinstance(field, Compound):
        temp_name = name
        if field.name:
            temp_name = f"{name}.{field.name}"

        out = []
        for sub_name, sub_field in field.fields().items():
            sub_name = f"{temp_name}.{sub_name}"
            out.extend(build_templates(sub_name, sub_field))

        return out

    elif isinstance(field, Optional):
        return build_templates(name, field.child_type, nested_template=nested_template)

    else:
        raise NotImplementedError(f"Unknown type for elasticsearch dynamic mapping: {field.__class__}")



"""Sequence manipulation methods used in parsing raw datastore output."""
from typing import Sequence, Generator, List


def chunk(items: Sequence, n: int) -> Generator:
    """ Yield n-sized chunks from list.

    >>> list(chunk([1,2,3,4,5,6,7], 2))
    [[1,2], [3,4], [5,6], [7,]]
    """
    for i in range(0, len(items), n):
        yield items[i:i+n]


def chunked_list(items: Sequence, n: int) -> List:
    """ Create a list of n-sized chunks from list.

    >>> chunked_list([1,2,3,4,5,6,7], 2)
    [[1,2], [3,4], [5,6], [7,]]
    """
    return list(chunk(items, n))



import ftplib
import logging
import os
import posixpath
import time
import errno

from io import BytesIO

from assemblyline.common.exceptions import ChainAll
from assemblyline.common.path import splitpath
from assemblyline.common.uid import get_random_id
from assemblyline.filestore.transport.base import Transport, TransportException, normalize_srl_path


def reconnect_retry_on_fail(func):
    def new_func(self, *args, **kwargs):
        max_retry = 3
        try_count = 0

        while try_count < max_retry:
            try:
                if not self.ftp:
                    if self.use_tls:
                        self.ftp = ftplib.FTP_TLS()
                        self.ftp.connect(self.host, port=self.port)
                        self.ftp.auth()
                        self.ftp.prot_p()

                    else:
                        self.ftp = ftplib.FTP()
                        self.ftp.connect(self.host, port=self.port)

                    self.ftp.login(self.user, self.password)
                    self.ftp.sendcmd("TYPE I")
                    try:
                        self.ftp.voidcmd('site umask 002')
                    except ftplib.Error:
                        pass

                return func(self, *args, **kwargs)
            except ftplib.error_perm as e:
                if str(e).startswith('550'):
                    raise
                msg = str(e) or "Unknown permission error"

            except ftplib.error_temp as e:
                msg = str(e) or "Unknown temporary error"

            except ftplib.error_reply as e:
                msg = str(e) or "Unknown reply error"

            except ftplib.Error as e:
                msg = str(e) or "Unknown FTP Error"

            except IOError as e:
                # May need to ignore other errors as well
                if e.errno not in (errno.EPIPE, errno.ECONNRESET):
                    raise
                msg = "IOError #%s" % e.errno

            # Prevent any stale connection errors to show up in the warnings
            # as these error can happen often if the ftp connection is not used
            # enough.
            if msg.startswith('421') or msg.startswith('425') or msg == "IOError #32" or msg == "IOError #104":
                self.log.info("FTP [%s]: %s" % (self.host, msg))
            else:
                self.log.warning("FTP [%s]: %s" % (self.host, msg))

            # The previous attempt at calling original func failed.
            # Reset the connection and try again.
            try:
                if self.ftp:
                    self.ftp.close()  # Just best effort.
            except ftplib.Error:
                pass

            time.sleep((2 ** try_count) / (2.00 ** (max_retry * 2)))

            self.ftp = None
            try_count += 1

        raise TransportException("Max retries reach for function: %s(%s)" % (func.__name__, ", ".join(map(repr, args))))

    new_func.__name__ = func.__name__
    new_func.__doc__ = func.__doc__
    return new_func


@ChainAll(TransportException)
class TransportFTP(Transport):
    """
    FTP Transport class.
    """
    def __init__(self, base=None, host=None, password=None, user=None, port=None, use_tls=None):
        self.log = logging.getLogger('assemblyline.transport.ftp')
        self.base = base
        self.ftp = None
        self.host = host
        self.port = int(port or 21)
        self.password = password
        self.user = user
        self.use_tls = use_tls

        def ftp_normalize(path):
            # If they've provided an absolute path. Leave it a is.
            if path.startswith('/'):
                s = path
            # Relative paths
            elif '/' in path or len(path) != 64:
                s = posixpath.join(self.base, path)
            else:
                s = posixpath.join(self.base, normalize_srl_path(path))
            self.log.debug('ftp normalized: %s -> %s', path, s)
            return s

        super(TransportFTP, self).__init__(normalize=ftp_normalize)

    def __str__(self):
        out = 'ftp://{}@{}'.format(self.user, self.host)
        if self.base:
            out += self.base
        return out

    def close(self):
        if self.ftp:
            self.ftp.close()

    @reconnect_retry_on_fail
    def delete(self, path):
        path = self.normalize(path)
        self.ftp.delete(path)

    @reconnect_retry_on_fail
    def exists(self, path):
        path = self.normalize(path)
        self.log.debug('Checking for existence of %s', path)
        size = None
        try:
            size = self.ftp.size(path)
        except ftplib.error_perm as e:
            # If the file doesnt exist we get a 550.
            if not str(e).startswith('550'):
                raise
        return size is not None

    @reconnect_retry_on_fail
    def makedirs(self, path):
        self.log.debug("making dirs: %s", path)
        subdirs = splitpath(path, '/')
        for i in range(len(subdirs)):
            try:
                d = posixpath.sep + posixpath.join(*subdirs[:i + 1])
                self.ftp.mkd(d)
            except ftplib.Error:
                pass

    # File based functions
    @reconnect_retry_on_fail
    def download(self, src_path, dst_path):
        dir_path = os.path.dirname(dst_path)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        with open(dst_path, 'wb') as localfile:
            src_path = self.normalize(src_path)
            self.ftp.retrbinary('RETR ' + src_path, localfile.write)

    @reconnect_retry_on_fail
    def upload(self, src_path, dst_path):
        dst_path = self.normalize(dst_path)
        dirname = posixpath.dirname(dst_path)
        filename = posixpath.basename(dst_path)
        tempname = get_random_id()
        temppath = posixpath.join(dirname, tempname)
        finalpath = posixpath.join(dirname, filename)
        assert (finalpath == dst_path)
        self.makedirs(dirname)
        with open(src_path, 'rb') as localfile:
            self.log.debug("Storing: %s", temppath)
            self.ftp.storbinary('STOR ' + temppath, localfile)
        self.log.debug("Rename: %s -> %s", temppath, finalpath)
        self.ftp.rename(temppath, finalpath)
        assert (self.exists(dst_path))

    @reconnect_retry_on_fail
    def upload_batch(self, local_remote_tuples):
        return super(TransportFTP, self).upload_batch(local_remote_tuples)

    # Buffer based functions
    @reconnect_retry_on_fail
    def get(self, path):
        path = self.normalize(path)
        bio = BytesIO()
        self.ftp.retrbinary('RETR ' + path, bio.write)
        return bio.getvalue()

    @reconnect_retry_on_fail
    def put(self, dst_path, content):
        dst_path = self.normalize(dst_path)
        dirname = posixpath.dirname(dst_path)
        filename = posixpath.basename(dst_path)
        tempname = get_random_id()
        temppath = posixpath.join(dirname, tempname)
        finalpath = posixpath.join(dirname, filename)
        assert (finalpath == dst_path)
        self.makedirs(dirname)

        if isinstance(content, str):
            content = content.encode('utf-8')

        with BytesIO(content) as file_io:
            self.log.debug("Storing: %s", temppath)
            self.ftp.storbinary('STOR ' + temppath, file_io)

            self.log.debug("Rename: %s -> %s", temppath, finalpath)
            self.ftp.rename(temppath, finalpath)
            assert (self.exists(dst_path))
